# /*##########################################################################
#
# Copyright (c) 2016-2024 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

"""This module initializes matplotlib and sets-up the backend to use.

It MUST be imported prior to any other import of matplotlib.

It provides the matplotlib :class:`FigureCanvasQTAgg` class corresponding
to the used backend.
"""

from __future__ import annotations


__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "12/03/2026"


import io
import matplotlib
import numpy

from qtpy import API as BINDING
from qtpy.QtGui import QFont

# This must be performed before any import from matplotlib
if BINDING in ("pyside6", "pyqt6"):
    matplotlib.use("QtAgg", force=False)
    from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg  # noqa
    from matplotlib.backends.backend_qtagg import NavigationToolbar2QT  # noqa
elif BINDING =="pyqt5":
    matplotlib.use("Qt5Agg", force=False)
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg  # noqa
    from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT  # noqa
else:
    raise ImportError("Unsupported Qt binding: %s" % BINDING)


from matplotlib.font_manager import FontProperties
from matplotlib.mathtext import MathTextParser
from matplotlib.ticker import ScalarFormatter as _ScalarFormatter
from matplotlib import figure, font_manager
from packaging.version import Version

_MATPLOTLIB_VERSION = Version(matplotlib.__version__)


class DefaultTickFormatter(_ScalarFormatter):
    """Tick label formatter"""

    def __init__(self):
        super().__init__(useOffset=True, useMathText=True)
        self.set_scientific(True)
        self.create_dummy_axis()

    if _MATPLOTLIB_VERSION < Version("3.1.0"):

        def format_ticks(self, values):
            self.set_locs(values)
            return [self(value, i) for i, value in enumerate(values)]


_FONT_STYLES = {
    QFont.StyleNormal: "normal",
    QFont.StyleItalic: "italic",
    QFont.StyleOblique: "oblique",
}


def qFontToFontProperties(font: QFont):
    """Convert a QFont to a matplotlib FontProperties"""
    weightFactor = 10 if BINDING == "PyQt5" else 1
    families = [font.family(), font.defaultFamily()]
    if _MATPLOTLIB_VERSION >= Version("3.6.0"):
        # Prevent 'Font family not found' warnings
        availableNames = font_manager.get_font_names()
        families = [f for f in families if f in availableNames]
        families.append(font_manager.fontManager.defaultFamily["ttf"])

    if "Sans" in font.family():
        families.insert(0, "sans-serif")

    return FontProperties(
        family=families,
        style=_FONT_STYLES[font.style()],
        weight=weightFactor * font.weight(),
        size=font.pointSizeF(),
    )


def rasterMathText(
    text: str,
    font: QFont,
    dotsPerInch: float = 96.0,
) -> tuple[numpy.ndarray, float]:
    """Raster text using matplotlib supporting latex-like math syntax.

    It supports multiple lines.

    :param text: The text to raster
    :param font: Font to use
    :param dotsPerInch: The DPI resolution of the created image
    :return: Corresponding image in gray scale and baseline offset from top
    """
    # Implementation adapted from:
    # https://github.com/matplotlib/matplotlib/blob/d624571a19aec7c7d4a24123643288fc27db17e7/lib/matplotlib/mathtext.py#L264

    stripped_text = text.strip("\n")
    font_prop = qFontToFontProperties(font)

    parser = MathTextParser("path")
    lines_info = [
        parser.parse(line, prop=font_prop, dpi=dotsPerInch)
        for line in stripped_text.split("\n")
    ]
    max_line_width = max(info[0] for info in lines_info)
    # Use lp string as minimum height/ascent
    ref_info = parser.parse("lp", prop=font_prop, dpi=dotsPerInch)
    line_height = max(
        ref_info[1],
        *(info[1] for info in lines_info),
    )
    first_line_ascent = max(
        ref_info[1] - ref_info[2], lines_info[0][1] - lines_info[0][2]
    )

    linespacing = 1.2

    figure_height = numpy.ceil(line_height * len(lines_info) * linespacing) + 2
    fig = figure.Figure(
        figsize=(
            (max_line_width + 1) / dotsPerInch,
            figure_height / dotsPerInch,
        )
    )
    fig.set_dpi(dotsPerInch)
    text = fig.text(
        0,
        1,
        stripped_text,
        fontproperties=font_prop,
        verticalalignment="top",
    )
    text.set_linespacing(linespacing)
    with io.BytesIO() as buffer:
        fig.savefig(buffer, dpi=dotsPerInch, format="raw")
        canvas_width, canvas_height = fig.get_window_extent().max
        buffer.seek(0)
        image = numpy.frombuffer(buffer.read(), dtype=numpy.uint8).reshape(
            int(canvas_height), int(canvas_width), 4
        )

    # RGB to inverted R channel
    array = 255 - image[:, :, 0]

    # Remove leading/trailing empty columns and trailing rows but one on each side
    filled_rows = numpy.nonzero(numpy.sum(array, axis=1))[0]
    filled_columns = numpy.nonzero(numpy.sum(array, axis=0))[0]
    if len(filled_rows) == 0 or len(filled_columns) == 0:
        return array, first_line_ascent
    return (
        numpy.ascontiguousarray(
            array[
                0 : filled_rows[-1] + 2,
                max(0, filled_columns[0] - 1) : filled_columns[-1] + 2,
            ]
        ),
        first_line_ascent,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
"""
Test to check TiffIO
"""

import unittest
import os
import logging
import numpy
from .utilstest import UtilsTest
from ..TiffIO import TiffIO

logger = logging.getLogger(__name__)


class TestTiffIO(unittest.TestCase):
    """Test the class format"""

    def write(self, filename):
        tif = TiffIO(filename, mode="wb+")
        dtype = numpy.uint16
        data = numpy.arange(10000).astype(dtype)
        data = data.reshape(100, 100)
        tif.writeImage(data, info={"Title": "1st"})
        tif = None

    def test_write(self):
        filename = "%s.tiff" % self.id()
        filename = os.path.join(UtilsTest.tempdir, filename)
        self.write(filename)

    def test_append(self):
        filename = "%s.tiff" % self.id()
        filename = os.path.join(UtilsTest.tempdir, filename)
        self.write(filename)
        # append
        tif = TiffIO(filename, mode="rb+")
        dtype = numpy.uint16
        data = numpy.arange(100).astype(dtype)
        data = data.reshape(10, 10)
        tif.writeImage((data * 2).astype(dtype), info={"Title": "2nd"})
        self.assertEqual(tif.getNumberOfImages(), 2)
        tif = None

    def test_read(self):
        filename = "%s.tiff" % self.id()
        filename = os.path.join(UtilsTest.tempdir, filename)
        self.write(filename)

        tif = TiffIO(filename)
        self.assertEqual(tif.getNumberOfImages(), 1)
        for i in range(tif.getNumberOfImages()):
            info = tif.getInfo(i)
            for key in info:
                if key not in ["colormap"]:
                    logger.info("%s = %s", key, info[key])
                elif info["colormap"] is not None:
                    logger.info("RED   %s = %s", key, info[key][0:10, 0])
                    logger.info("GREEN %s = %s", key, info[key][0:10, 1])
                    logger.info("BLUE  %s = %s", key, info[key][0:10, 2])
            data = tif.getImage(i)[0, 0:10]
            logger.info("data [0, 0:10] = %s", data)


def suite():
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loader(TestTiffIO))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python

__date__ = "27/10/2025"
__author__ = "Jérôme Kieffer"
__license__ = "MIT"

import os
import sys
import argparse
from io import StringIO
import logging
import numpy
from .. import version as fabio_version, date as fabio_date
from ..openimage import openimage as fabio_open
from ..nexus import Nexus
import h5py
import hdf5plugin

application_name = os.path.splitext(os.path.basename(__file__))[0]
logger = logging.getLogger(application_name)
logging.basicConfig()

try:
    from pyFAI import load
except ImportError:
    pyFAI = None
    logger.error("Unable to import pyFAI, won't be able to parse PONI-file!")


class XDSbuilder:
    def __init__(self):
        "Constructor of the class"

        self.poni = None
        self.options = None
        self.frames = None
        self.h5_filename = "fabio_master.h5"

    def parse(self, argv=None):
        "Parse command line arguments and return"
        if argv is None:
            argv = []
        parser = argparse.ArgumentParser(
            prog=application_name,
            description="Convert any HDF5 file containing images to a file processable by XDS using the neggia plugin from Dectris."
            "Do not forget to specify LIB=/path/to/plugin/dectris-neggia.so in XDS.INP",
            epilog="Requires hdf5plugin and pyFAI to parse configuration file. Geometry can be calibrated with pyFAI-calib2",
        )
        parser.add_argument(
            "--version",
            action="version",
            version=f"%(prog)s {fabio_version} from {fabio_date}",
        )
        parser.add_argument(
            "input", nargs="+", help="Space separated list of input files (HDF5)"
        )
        parser.add_argument(
            "--verbose", "-v", help="increase output verbosity", action="count"
        )
        parser.add_argument(
            "--force", "-f", help="force overwrite output file", action="store_true"
        )
        parser.add_argument(
            "--copy",
            "-c",
            help="copy dataset instead of using external links",
            action="store_true",
        )
        parser.add_argument(
            "--geometry",
            "-g",
            help="PONI-file containing the geometry (pyFAI format, MANDATORY)",
        )
        parser.add_argument(
            "--output",
            "-o",
            help=f"output directory, the data will be in {self.h5_filename}",
            default="fabio_xds",
        )
        parser.add_argument(
            "--CdTe",
            help="The detector is made of CdTe",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--neggia", help="Path of the neggia plugin", default="dectris-neggia.so"
        )
        parser.add_argument(
            "--oscillation", help="Oscillation range used.", default=0.5
        )

        self.options = parser.parse_args(argv)
        return self.options

    def configure_verboseness(self):
        if self.options and self.options.verbose:
            if self.options.verbose > 1:
                logger.setLevel(logging.DEBUG)
            else:
                logger.setLevel(logging.INFO)

    def load_poni(self):
        "return 1 if poni-file cannot be found"
        if self.options is None:
            return 1
        if not self.options.geometry or not os.path.exists(self.options.geometry):
            logger.error("Unable to parse PONI-file: %s", self.options.geometry)
            return 1
        try:
            self.poni = load(self.options.geometry)
        except Exception as err:
            logger.error("Unable to parse PONI-file: %s", self.options.geometry)
            raise err
            return 1

    def load_input(self):
        if len(self.options.input) == 0:
            logger.error("No input HDF5 file provided. Aborting")
            return 2
        self.frames = [fabio_open(i) for i in self.options.input]
        return 0

    def _distance_center(self):
        """return distance, centerx, center_y"""
        width = self.poni.detector.shape[-1]
        distance = self.poni.dist * 1e3  # mm
        r_array = self.poni.array_from_unit(unit="r_m")
        x_center = numpy.argmin(r_array) % width
        y_center = numpy.argmin(r_array) // width
        return distance, x_center, y_center

    def build_neggia(self):
        """Build the neggia file, i.e. the HDF5 file with data + metadata for analysis"""
        if not os.path.exists(self.options.output):
            os.makedirs(self.options.output)
        dest_h5 = os.path.join(self.options.output, self.h5_filename)
        if os.path.exists(dest_h5):
            if self.options.force:
                mode = "w"
            else:
                logger.error("Output file exist, not overwriting it. Aborting")
                return 1
        else:
            mode = "w"
        distance, center_x, center_y = self._distance_center()
        with Nexus(dest_h5, mode) as nxs:
            entry = nxs.new_entry(
                entry="entry", program_name=application_name, force_name=True
            )
            instrument = nxs.new_instrument(entry=entry, instrument_name="instrument")
            if self.poni.wavelength:
                beam = nxs.new_class(instrument, "beam", "NXbeam")
                beam.create_dataset(
                    "incident_wavelength", data=self.poni.wavelength * 1e10
                ).attrs["unit"] = "A"
            detector = nxs.new_class(instrument, "detector", "NXdetector")
            detector.create_dataset("x_pixel_size", data=float(self.poni.pixel2)).attrs[
                "unit"
            ] = "m"
            detector.create_dataset("y_pixel_size", data=float(self.poni.pixel1)).attrs[
                "unit"
            ] = "m"
            detector.create_dataset("beam_center_x", data=float(center_x)).attrs[
                "unit"
            ] = "pixel"
            detector.create_dataset("beam_center_y", data=float(center_y)).attrs[
                "unit"
            ] = "pixel"
            detector.create_dataset("detector_distance", data=distance * 1e-3).attrs[
                "unit"
            ] = "m"
            detectorSpecific = nxs.new_class(
                detector, "detectorSpecific", "NXcollection"
            )
            detectorSpecific.create_dataset(
                "nimages", data=sum(i.nframes for i in self.frames)
            )
            detectorSpecific.create_dataset("ntrigger", data=1)
            mask = self.poni.detector.mask
            if mask is None:
                mask = numpy.zeros(self.poni.detector.shape, dtype="uint32")
            else:
                mask = mask.astype("uint32")
            detectorSpecific.create_dataset("pixel_mask", data=mask)
            data = nxs.new_class(entry, "data", "NXdata")
            cnt = 0
            for fimg in self.frames:
                if isinstance(fimg.dataset, h5py.Dataset):
                    cnt += 1
                    if self.options.copy:
                        data.create_dataset(
                            f"data_{cnt:06d}",
                            data=fimg.dataset[()],
                            chunks=(1,) + fimg.shape,
                            **hdf5plugin.Bitshuffle(nelems=0, cname="lz4"),
                        )
                    else:
                        data[f"data_{cnt:06d}"] = h5py.ExternalLink(
                            os.path.relpath(
                                fimg.dataset.file.filename, self.options.output
                            ),
                            fimg.dataset.name,
                        )
                elif isinstance(fimg.dataset, numpy.ndarray):
                    cnt += 1
                    if fimg.dataset.ndim < 3:
                        dataset = numpy.atleast_3d(fimg.dataset)
                        new_shape = (1,) * (3 - fimg.dataset.ndim) + fimg.dataset.shape
                        dataset = dataset.reshape(new_shape)
                    else:
                        dataset = fimg.dataset
                    data.create_dataset(
                        f"data_{cnt:06d}",
                        data=dataset,
                        chunks=(1,) + fimg.shape,
                        **hdf5plugin.Bitshuffle(nelems=0, cname="lz4"),
                    )
                else:  # assume it is a list
                    for item in fimg.dataset:
                        # each item can be either a dataset or a numpy array
                        if isinstance(item, h5py.Dataset):
                            cnt += 1
                            if self.options.copy:
                                data.create_dataset(
                                    f"data_{cnt:06d}",
                                    data=item[()],
                                    chunks=(1,) + fimg.shape,
                                    **hdf5plugin.Bitshuffle(nelems=0, cname="lz4"),
                                )
                            else:
                                data[f"data_{cnt:06d}"] = h5py.ExternalLink(
                                    os.path.relpath(
                                        item.file.filename, self.options.output
                                    ),
                                    item.name,
                                )

                        elif isinstance(fimg.dataset, numpy.ndarray):
                            cnt += 1
                            if fimg.dataset.ndim < 3:
                                dataset = numpy.atleast_3d(item)
                                new_shape = (1,) * (3 - item.ndim) + item.shape
                                dataset = dataset.reshape(new_shape)
                            else:
                                dataset = item
                            data.create_dataset(
                                f"data_{cnt:06d}",
                                data=dataset,
                                chunks=(1,) + fimg.shape,
                                **hdf5plugin.Bitshuffle(nelems=0, cname="lz4"),
                            )
                        else:
                            logger.warning(
                                "Don't know how to handle %s, skipping", item
                            )
        os.link(dest_h5, dest_h5.replace("master", "000000"))
        logger.info(f"Successfully created Neggia file {dest_h5}.")
        return 0

    def build_XDS(self):
        "Create XDS.INP file suitable for data reduction"
        if not os.path.exists(self.options.output):
            os.makedirs(self.options.output)
        dest_xds = os.path.join(self.options.output, "XDS.INP")

        xds = [
            "JOB= XYCORR INIT COLSPOT IDXREF DEFPIX INTEGRATE CORRECT",
            "! JOB=  DEFPIX INTEGRATE CORRECT",
            "",
            f"NAME_TEMPLATE_OF_DATA_FRAMES= {self.h5_filename.replace('master', '??????')}",
            f"LIB= {self.options.neggia}",
            "",
            "SPACE_GROUP_NUMBER=0 ! 0 if unknown",
            "FRIEDEL'S_LAW=FALSE     ! This acts only on the CORRECT step",
            "",
            "FRACTION_OF_POLARIZATION=0.99",
            "ROTATION_AXIS=0 -1  0",
            "POLARIZATION_PLANE_NORMAL=0 1 0",
            "DIRECTION_OF_DETECTOR_X-AXIS=1 0 0",
            "DIRECTION_OF_DETECTOR_Y-AXIS=0 1 0",
            "INCIDENT_BEAM_DIRECTION=0 0 1",
            f"OSCILLATION_RANGE= {self.options.oscillation}",
            "",
            "OVERLOAD=100000000",
        ]
        xds.append(f"DATA_RANGE= 1 {sum(i.nframes for i in self.frames)}")

        shape = self.poni.detector.shape
        pixel1, pixel2 = self.poni.detector.pixel1, self.poni.detector.pixel2
        xds.append(
            f"NX= {shape[1]:d} NY= {shape[0]:d}  QX= {pixel2 * 1000:f}  QY= {pixel1 * 1000:f}"
        )

        distance, center_x, center_y = self._distance_center()
        xds.append("DETECTOR= EIGER")
        xds.append(f"DETECTOR_DISTANCE= {distance:.4f}")
        xds.append(f"ORGX= {center_x:.3f} ORGY= {center_y:.3f}")
        xds.append(f"X-RAY_WAVELENGTH={1e10 * self.poni.wavelength}")
        xds.append("")
        mask = self.poni.detector.mask > 0
        empty = numpy.where(numpy.std(mask, axis=0) == 0)[0]
        if len(empty):
            start = 0
            for stop in numpy.where(empty[1:] - empty[:-1] != 1)[0]:
                xds.append(
                    f"UNTRUSTED_RECTANGLE= {empty[start]:4d} {empty[stop]:4d}    0 {shape[0]:4d} "
                )
                start = stop + 1
            xds.append(
                f"UNTRUSTED_RECTANGLE= {empty[start]:4d} {empty[-1]:4d}    0 {shape[0]:4d} "
            )

        empty = numpy.where(numpy.std(mask, axis=1) == 0)[0]
        if len(empty):
            start = 0
            for stop in numpy.where(empty[1:] - empty[:-1] != 1)[0]:
                xds.append(
                    f"UNTRUSTED_RECTANGLE=     0 {shape[1]:4d} {empty[start]:4d} {empty[stop]:4d}"
                )
                start = stop + 1
            xds.append(
                f"UNTRUSTED_RECTANGLE=     0 {shape[1]:4d} {empty[start]:4d} {empty[-1]:4d}"
            )
        xds.append("")
        if self.options.CdTe:
            xds.append("SENSOR_THICKNESS=0.75 !mm")
            nrj = self.poni.energy
            e, mu = numpy.loadtxt(StringIO(Attenuations_CdTe), unpack=True)
            xds.append(f"SILICON= {numpy.interp(nrj, e, mu):.3f} !1/mm")

        mode = "w"
        if os.path.exists(dest_xds):
            if self.options.force:
                mode = "w"
            else:
                logger.error("Output file exist, not overwriting it. Aborting")
                return 1
        with open(dest_xds, mode) as w:
            w.write(os.linesep.join(xds))
        logger.info(f"Successfully created XDS file {dest_xds}.")
        return 0

    def process(self):
        "pipeline processing, returns a error code as integer"
        self.configure_verboseness()

        rc = self.load_poni()
        if rc:
            return rc

        rc = self.load_input()
        if rc:
            return rc

        rc = self.build_neggia()
        if rc:
            return rc

        return self.build_XDS()


def main(argv=None):
    xds = XDSbuilder()
    xds.parse(sys.argv[1:] if argv is None else argv)
    return xds.process()


if __name__ == "__main__":
    sys.exit(main())


Attenuations_CdTe = """# E(keV)    µ(1/mm)
  5    4.8824e+02
  6    3.0683e+02
  7    2.0446e+02
  8    1.4391e+02
  9    1.0501e+02
 10    7.9268e+01
 11    6.1074e+01
 12    4.8186e+01
 13    3.8739e+01
 14    3.1654e+01
 15    2.6231e+01
 16    2.1938e+01
 17    1.8545e+01
 18    1.5830e+01
 19    1.3636e+01
 20    1.1835e+01
 21    1.0319e+01
 22    9.0617e+00
 23    8.0028e+00
 24    7.1078e+00
 25    6.3414e+00
 26    5.6880e+00
 27    1.6275e+01
 28    1.4806e+01
 29    1.3508e+01
 30    1.2367e+01
 31    1.1343e+01
 32    1.9767e+01
 33    1.8252e+01
 34    1.6895e+01
 35    1.5666e+01
 36    1.4555e+01
 37    1.3549e+01
 38    1.2630e+01
 39    1.1794e+01
 40    1.1027e+01
 41    1.0331e+01
 42    9.6876e+00
 43    9.0968e+00
 44    8.5527e+00
 45    8.0496e+00
 46    7.5875e+00
 47    7.1604e+00
 48    6.7626e+00
 49    6.3940e+00
 50    6.0547e+00
 51    5.7383e+00
 52    5.4440e+00
 53    5.1696e+00
 54    4.9140e+00
 55    4.6753e+00
 56    4.4518e+00
 57    4.2424e+00
 58    4.0464e+00
 59    3.8622e+00
 60    3.6896e+00
 61    3.5270e+00
 62    3.3737e+00
 63    3.2298e+00
 64    3.0935e+00
 65    2.9654e+00
 66    2.8443e+00
 67    2.7296e+00
 68    2.6214e+00
 69    2.5184e+00
 70    2.4213e+00
 71    2.3295e+00
 72    2.2423e+00
 73    2.1592e+00
 74    2.0803e+00
 75    2.0054e+00
 76    1.9340e+00
 77    1.8662e+00
 78    1.8018e+00
 79    1.7404e+00
 80    1.6819e+00
 81    1.6257e+00
 82    1.5725e+00
 83    1.5210e+00
 84    1.4724e+00
 85    1.4262e+00
 86    1.3812e+00
 87    1.3391e+00
 88    1.2981e+00
 89    1.2589e+00
 90    1.2215e+00
 91    1.1858e+00
 92    1.1513e+00
 93    1.1179e+00
 94    1.0863e+00
 95    1.0559e+00
 96    1.0267e+00
 97    9.9859e-01
 98    9.7168e-01
 99    9.4536e-01
100    9.2021e-01
"""



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation files
#  (the "Software"), to deal in the Software without restriction,
#  including without limitation the rights to use, copy, modify, merge,
#  publish, distribute, sublicense, and/or sell copies of the Software,
#  and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION W

"""
Authors: Henning O. Sorensen & Erik Knudsen
         Center for Fundamental Research: Metal Structures in Four Dimensions
         Risoe National Laboratory
         Frederiksborgvej 399
         DK-4000 Roskilde
         email:erik.knudsen@risoe.dk

        + Jon Wright, ESRF
"""

import logging
import numpy
from .fabioimage import FabioImage
from .fabioutils import ENDIANNESS

logger = logging.getLogger(__name__)

DATA_TYPES = {
    2: numpy.int16,
    4: numpy.uint16,
    3: numpy.int32,
    5: numpy.uint32,
    6: numpy.float32,
    7: numpy.float64,
    8: numpy.int8,
    9: None,
    10: None,
    15: "Struct",
    18: None,
    20: None,
}

DATA_BYTES = {
    2: 2,
    4: 2,
    3: 4,
    5: 4,
    6: 4,
    7: 8,
    8: 1,
    9: None,
    10: None,
    15: "Struct",
    18: None,
    20: None,
}

BE_uint16 = numpy.dtype("uint16").newbyteorder(">")
BE_uint32 = numpy.dtype("uint32").newbyteorder(">")
BE_int32 = numpy.dtype("int32").newbyteorder(">")

class Dm3Image(FabioImage):
    """Read and try to write the dm3 data format"""

    DESCRIPTION = "Digital Micrograph DM3 file format"

    DEFAULT_EXTENSIONS = ["dm3"]

    def __init__(self, *args, **kwargs):
        FabioImage.__init__(self, *args, **kwargs)
        self.encoded_datatype = None
        self.no_data_elements = None
        self.grouptag_is_sorted = None
        self.grouptag_is_open = None
        self.tag_encoded_type = None
        self.tag_data_type = None
        self.tag_is_data = None
        self.grouptag_no_tags = None
        self.bytes_in_file = None
        self.tag_label_length = None
        self.byte_order = None

    def _readheader(self):
        self.infile.seek(0)
        file_format = self.readbytes(4, BE_uint32)[0]  # should be 3
        if file_format != 3:
            raise RuntimeError("Wrong file type")
        self.bytes_in_file = self.readbytes(4, BE_uint32)[0]
        byte_order = self.readbytes(4, BE_uint32)[0]  # 0 = big, 1= little
        logger.debug("read dm3 file - file format %s" % file_format)
        logger.debug("Bytes in file: %s" % self.bytes_in_file)
        logger.debug("Byte order: %s  - 0 = bigEndian , 1 = littleEndian", self.byte_order)

        if byte_order == 0:
            self.byte_order = ENDIANNESS.BIG
        elif byte_order == 1:
            self.byte_order = ENDIANNESS.LITTLE
        else:
            raise ValueError("Inconsistent endianness in file")

    def read(self, fname, frame=None):
        self.header = self.check_header()
        self.resetvals()
        self.infile = self._open(fname, "rb")
        self._readheader()
        go_on = True
        while go_on:
            self.read_tag_group()
            self.read_tag_entry()
            if self.infile.tell() > self.bytes_in_file:
                break

            while self.tag_is_data == 21:
                self.read_tag_entry()
                if self.infile.tell() > self.bytes_in_file:
                    go_on = False

        dim_raw = self.header["Active Size (pixels)"].split()
        dim1_raw = int(dim_raw[0])
        dim2_raw = int(dim_raw[1])
        binning_raw = self.header["Binning"]
        try:
            dim1_binning, dim2_binning = map(int, binning_raw.split())
        except AttributeError:
            dim1_binning, dim2_binning = map(lambda x: x * int(binning_raw) * x, (1, 1))
        self._shape = dim2_raw // dim2_binning, dim1_raw // dim1_binning
        if "Data" in self.header:
            self.data = self.header["Data"].reshape(self._shape)
            self._shape = None
        return self

    def readbytes(self, bytes_to_read, format):
        raw = self.infile.read(bytes_to_read)
        if format is not None:
            data = numpy.frombuffer(raw, format).copy()
        else:
            data = raw
        return data

    def read_tag_group(self):
        self.grouptag_is_sorted = self.readbytes(1, numpy.uint8)[0]
        self.grouptag_is_open = self.readbytes(1, numpy.uint8)[0]
        self.grouptag_no_tags = self.readbytes(4, BE_uint32)[0]
        logger.debug("TagGroup is sorted? %s", self.grouptag_is_sorted)
        logger.debug("TagGroup is open? %s", self.grouptag_is_open)
        logger.debug("no of tags in TagGroup %s", self.grouptag_no_tags)

    def read_tag_entry(self):
        self.tag_is_data = self.readbytes(1, numpy.uint8)[0]
        self.tag_label_length = self.readbytes(2, BE_uint16)[0]
        logger.debug(
            "does Tag have data ? %s  -  20 = Tag group , 21 = data ", self.tag_is_data
        )
        logger.debug("length of tag_label %s", self.tag_label_length)
        if self.tag_label_length != 0:
            tag_label = self.infile.read(self.tag_label_length)
        else:
            tag_label = b""

        if self.tag_is_data == 21:
            # This is data
            try:
                key = tag_label.decode("latin-1")
            except Exception:
                key = tag_label.decode("latin-1", "replace")
                logger.warning("Non-valid latin-1 key renamed into '%s'" % key)
            value = self.read_tag_type()
            if isinstance(value, bytes):
                value = value.decode()
            logger.debug("%s: %s", key, value)
            if key in self.header:
                logger.debug(
                    "Key '%s' already exists with value %s. Overwritten with %s.",
                    key,
                    self.header[key],
                    value,
                )
            self.header[key] = value

    def read_tag_type(self):
        read = self.infile.read(4)
        if read != b"%%%%":
            raise IOError(f"Inconsistent file reading {read}")
        self.tag_data_type = self.readbytes(4, BE_uint32)[0]
        logger.debug(
            "data is of type: %s - 1 = simple, 2 = string, 3 = array, >3 structs.",
            self.tag_data_type)
        self.tag_encoded_type = self.readbytes(4, BE_uint32)[0]
        logger.debug(
            "encode type: %s %s",
            self.tag_encoded_type,
            DATA_TYPES[self.tag_encoded_type],
        )
        if self.tag_data_type == 1:
            # simple type
            return self.readbytes(
                DATA_BYTES[self.tag_encoded_type],
                self.get_stype(DATA_TYPES[self.tag_encoded_type], self.byte_order),
            )[0]
        # are the data stored in a simple array?
        if self.tag_encoded_type == 20 and self.tag_data_type == 3:
            self.data_type = self.readbytes(4, BE_uint32)[0]
            self.no_data_elements = self.readbytes(4, BE_uint32)[0]
            if self.data_type == 10:
                logger.debug("skip bytes %s", self.no_data_elements)
                _dump = self.infile.read(self.no_data_elements)
                return None

            logger.debug("Data are stored as a simple a array -")
            logger.debug(
                "%s data elements stored as %s", self.no_data_elements, self.data_type
            )
            read_no_bytes = DATA_BYTES[self.data_type] * self.no_data_elements
            fmt = self.get_stype(DATA_TYPES[self.data_type], self.byte_order)
            return self.readbytes(read_no_bytes, fmt)

        # are the data stored in a complex array ?
        # print 'tag_type + data_type', self.tag_encoded_type,self.tag_data_type

        # print self.tag_encoded_type , self.tag_data_type
        if self.tag_encoded_type == 20 and self.tag_data_type > 3:
            self.tag_encoded_type = self.readbytes(4, BE_uint32)[0]
            logger.debug("found array - new tag_encoded_type %s", self.tag_encoded_type)
            if self.tag_encoded_type == 15:  # struct type
                # ##type = self.readbytes(4,numpy.int32)
                _struct_name_length = self.readbytes(4, BE_int32)[0]
                struct_number_fields = self.readbytes(4, BE_int32)[0]
                # print 'struct - name_length, number_field',  struct_name_length,struct_number_fields
                # print self.infile.read(_struct_name_length)
                field_info = []
                for i in range(struct_number_fields):
                    field_info.append(
                        [
                            self.readbytes(4, BE_int32)[0],
                            self.readbytes(4, BE_int32)[0],
                        ]
                    )
                # print field_info
                self.no_data_elements = self.readbytes(4, BE_int32)[0]
                # print '%i data elemets stored as ' %self.no_data_elements
                bytes_in_struct = 0
                for i in range(struct_number_fields):
                    bytes_in_struct += DATA_BYTES[field_info[i][1]]
                logger.debug("skip bytes %s", self.no_data_elements * bytes_in_struct)
                _dump = self.infile.read(self.no_data_elements * bytes_in_struct)
                return None

        if self.tag_encoded_type == 15:  # struct type
            # ##type = self.readbytes(4,numpy.int32)
            _struct_name_length = self.readbytes(4, BE_int32)[0]
            struct_number_fields = self.readbytes(4, BE_int32)[0]
            # print 'struct - name_length, number_field', _struct_name_length,struct_number_fields
            # print self.infile.read(struct_name_length)
            field_info = []
            for i in range(struct_number_fields):
                field_info.append(
                    [
                        self.readbytes(4, BE_int32)[0],
                        self.readbytes(4, BE_int32)[0],
                    ]
                )
            # print field_info
            field_data = b""
            for i in range(struct_number_fields):
                field_data += self.readbytes(field_info[i][0], None) + b" "
                data = self.readbytes(
                    DATA_BYTES[field_info[i][1]],
                    self.get_stype(DATA_TYPES[field_info[i][1]], self.byte_order))
                field_data += str(data[0]).encode() + b" "
            return field_data

    def read_data(self):
        self.encoded_datatype = (
            numpy.frombuffer(self.infile.read(4), BE_uint32).copy()
        )


dm3image = Dm3Image



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

"""
Test cases for filename deconstruction


testsuite by Jerome Kieffer (Jerome.Kieffer@esrf.eu)
28/11/2014
"""

import unittest
import os
import logging
import tempfile
import fabio
import numpy

logger = logging.getLogger(__name__)

CASES = [
    (1, "edf", "data0001.edf"),
    (10001, "edf", "data10001.edf"),
    (10001, "edf", "data10001.edf.gz"),
    (10001, "edf", "data10001.edf.bz2"),
    (2, "marccd", "data0002.mccd"),
    (12345, "marccd", "data12345.mccd"),
    (10001, "marccd", "data10001.mccd.gz"),
    (10001, "marccd", "data10001.mccd.bz2"),
    (123, "marccd", "data123.mccd.gz"),
    (3, "tif_or_pilatus", "data0003.tif"),
    (4, "tif_or_pilatus", "data0004.tiff"),
    (12, "bruker", "sucrose101.012.gz"),
    (99, "bruker", "sucrose101.099"),
    (99, "bruker", "sucrose101.0099"),
    (99, "bruker", "sucrose101.0099.bz2"),
    (99, "bruker", "sucrose101.0099.gz"),
    (2, "fit2dmask", "fit2d.msk"),
    (None, "fit2dmask", "mymask.msk"),
    (670005, "edf", "S82P670005.edf"),
    (670005, "edf", "S82P670005.edf.gz"),
    # based on only the name it can be either img or oxd
    (1, "dtrek_or_oxd_or_hipic_or_raxis", "mb_LP_1_001.img"),
    (2, "dtrek_or_oxd_or_hipic_or_raxis", "mb_LP_1_002.img.gz"),
    (3, "dtrek_or_oxd_or_hipic_or_raxis", "mb_LP_1_003.img.bz2"),
    (3, "dtrek_or_oxd_or_hipic_or_raxis", os.path.join("data", "mb_LP_1_003.img.bz2")),
]

MORE_CASES = [
    ("data0010.edf", "data0012.edf", 10),
    ("data1000.pnm", "data999.pnm", 1000),
    ("data0999.pnm", "data1000.pnm", 999),
    ("data123457.edf", "data123456.edf", 123457),
    ("d0ata000100.mccd", "d0ata000012.mccd", 100),
    (
        os.path.join("images/sampledir", "P33S670003.edf"),
        os.path.join("images/sampledir", "P33S670002.edf"),
        670003,
    ),
    (
        os.path.join("images/P33S67", "P33S670003.edf"),
        os.path.join("images/P33S67", "P33S670002.edf"),
        670003,
    ),
    ("image2301.mar2300", "image2300.mar2300", 2301),
    ("image2300.mar2300", "image2301.mar2300", 2300),
    ("image.0123", "image.1234", 123),
    ("mymask.msk", "mymask.msk", None),
    ("data_123.mccd.bz2", "data_001.mccd.bz2", 123),
]


class TestFilenames(unittest.TestCase):
    """check the name -> number, type conversions"""

    def test_many_cases(self):
        """loop over CASES"""
        for num, typ, name in CASES:
            obj = fabio.FilenameObject(filename=name)
            self.assertEqual(
                num, obj.num, name + " num=" + str(num) + " != obj.num=" + str(obj.num)
            )
            self.assertEqual(
                typ, "_or_".join(obj.format), name + " " + "_or_".join(obj.format)
            )
            self.assertEqual(name, obj.tostring(), name + " " + obj.tostring())

    def test_more_cases(self):
        for nname, oname, num in MORE_CASES:
            name = fabio.construct_filename(oname, num)
            self.assertEqual(name, nname)

    def test_more_cases_jump(self):
        for nname, oname, num in MORE_CASES:
            name = fabio.jump_filename(oname, num)
            self.assertEqual(name, nname)


class TestFilenameObjects(unittest.TestCase):
    def setUp(self):
        """make a small test dataset"""
        self.datashape = (10, 11)
        self.nframes = 5
        self.tempdir = tempfile.mkdtemp()
        self.fnames = [
            os.path.join(self.tempdir, "FNO%04d.edf" % iframe)
            for iframe in range(self.nframes)
        ]
        data = numpy.zeros(self.datashape, numpy.uint16)
        im = fabio.edfimage.edfimage(data)
        for iframe, fname in enumerate(self.fnames):
            im.header["checkthing"] = str(iframe)
            im.write(fname)

    def tearDown(self):
        for name in self.fnames:
            os.remove(name)
        os.rmdir(self.tempdir)

    def test_files_are_being_opened(self):
        """Regression test for Fable"""
        for iframe, fname in enumerate(self.fnames):
            obj = fabio.FilenameObject(filename=fname)
            # read via the FilenameObject
            o1 = fabio.open(obj)
            self.assertEqual(o1.data.shape, self.datashape)
            self.assertEqual(o1.header["checkthing"], str(iframe))
            # And via the tostring
            o2 = fabio.open(obj.tostring())
            self.assertEqual(o2.data.shape, self.datashape)
            self.assertEqual(o2.header["checkthing"], str(iframe))

    def test_FileNameObject_can_iterate(self):
        """Regression test for Fable"""
        obj = fabio.FilenameObject(filename=self.fnames[0])
        for iframe, fname in enumerate(self.fnames):
            obj.num = iframe
            # read via the FilenameObject
            o1 = fabio.open(obj)
            self.assertEqual(o1.data.shape, self.datashape)
            self.assertEqual(o1.header["checkthing"], str(iframe))
            # And via the tostring
            o2 = fabio.open(obj.tostring())
            self.assertEqual(o2.data.shape, self.datashape)
            self.assertEqual(o2.header["checkthing"], str(iframe))
            # And the real name
            o3 = fabio.open(fname)
            self.assertEqual(o3.data.shape, self.datashape)
            self.assertEqual(o3.header["checkthing"], str(iframe))


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestFilenames))
    testsuite.addTest(loadTests(TestFilenameObjects))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fast Azimuthal integration
#             https://github.com/silx-kit/pyFAI
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Authors: Gael Goret <gael.goret@esrf.fr>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
"""
Portable diffraction images viewer/converter

* Written in Python, it combines the functionalities of the I/O library fabIO
  with a user friendly Qt GUI.
* Image converter is also a light viewer based on the visualization tool
  provided by the module matplotlib.
"""

__version__ = "1.1"
__author__ = "Gaël Goret, Jérôme Kieffer"
__copyright__ = "2015-2026 ESRF"
__licence__ = "MIT"

import os
import numpy
import fabio
from fabio.nexus import Nexus
# ----------------------------------------------------------------------
# Qt imports via QtPy – this works with PyQt5, PySide2, PySide6, etc.
# ----------------------------------------------------------------------
from qtpy import QtWidgets as qt
from qtpy import QtCore as qtc
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import (
    QSizePolicy,
    QFileDialog,
    QMessageBox,
    QAction,
    QComboBox,
    QPlainTextEdit,
    QLabel,
    QSplitter,
    QTabWidget,
    QWidget,
    QProgressBar,
    QGroupBox,
    QVBoxLayout,
    QHBoxLayout,
    QPushButton,
    QListWidget,
    QCheckBox,
    QButtonGroup
)
# Matplotlib imports (unchanged)
from matplotlib.figure import Figure
from .matplotlib import FigureCanvasQTAgg
from .matplotlib import NavigationToolbar2QT
from .dialogs import CounterFormatOptionDialog, DownSamplingDialog, BinDialog
# ----------------------------------------------------------------------
# Global configuration
# ----------------------------------------------------------------------
numpy.seterr(divide="ignore")

output_format = [
    "*.bin",
    "*.cbf",
    "*.edf",
    "*.h5",
    "*.img",
    "*.mar2300",
    "*.mar3450",
    "*.marccd",
    "*.tiff",
    "*.sfrm",
]
# ----------------------------------------------------------------------
# Main application class
# ----------------------------------------------------------------------
class AppForm(qt.QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)
        # ------------------------------------------------------------------
        # Window set‑up
        # ------------------------------------------------------------------
        self.setWindowTitle("FabIO Viewer")
        self.setSizePolicy(
            QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        )
        # Menu / widgets
        self.create_menu()
        self.create_main_frame()
        self.create_status_bar()
        # ------------------------------------------------------------------
        # Data containers
        # ------------------------------------------------------------------
        self.data = numpy.array([])
        self.header = []

        # Data Series
        self.imgDict = {}
        self.data_series = []
        self.header_series = []
        self.sequential_file_list = []
        self.sequential_file_dict = {}

        # Miscellaneous
        self.mask = None
        self.transform_data_series = False
        self.transform_list = []
        self.sequential_file_mode = False
        self.h5_loaded = False
        self.counter_format = "%03d"
    # ----------------------------------------------------------------------
    # Helper methods
    # ----------------------------------------------------------------------
    def format_header(self, d):
        """Return a nicely formatted header string.

        :param d: dict containing headers
        :return: formatted string
	"""
        keys = sorted(d.keys())
        return " \n".join([f"{k}: {d[k]}" for k in keys]) + " \n"

    def _open(self, filename):
        """Open an image with fabio, falling back to manual binary import."""
        try:
            img = fabio.open(filename)
        except Exception:
            message = (
                "Automatic format recognition procedure failed or perhaps you are "
                "trying to open a binary data block...\n\nSwitch to manual procedure."
            )
            QMessageBox.warning(self, "Message", message)
            dim1, dim2, offset, bytecode, endian = BinDialog(self).exec_()
            if dim1 is not None and dim2 is not None:
                endian = "<" if endian == "Short" else ">"
                img = fabio.binaryimage.binaryimage()
                img.read(filename, dim1, dim2, offset, bytecode, endian)
                img.header = {
                    "Info": "No header information available in binary data blocks"
                }
            else:
                return None
        return img
    # ----------------------------------------------------------------------
    # File‑opening / series handling
    # ----------------------------------------------------------------------
    def open_data_series(self, series=None):
        if not series:
            series = QFileDialog.getOpenFileNames(
                self, "Select and open series of files"
            )
            if isinstance(series, tuple):
                # PyQt5 compatibility – QtPy returns a tuple as well
                series = series[0]
        series = [str(f) for f in list(series)]
        total = len(series)
        if total == 0:
            return
        # Reset UI / internal state
        self.data_series = []
        self.header_series = []
        self.sequential_file_list = []
        self.imgDict = {}
        self.sequential_file_dict = {}
        self.images_list.clear()
        self.imagelistWidget.clear()
        self.headerTextEdit.clear()
        self.axes.clear()
        self.canvas.draw()
        self.h5_loaded = False
        iid = 0
        for fname in series:
            if not fname:
                continue
            extract_fname = self.extract_fname_from_path(fname)
            if self.sequential_file_mode:
                self.statusBar().showMessage(
                    f"Adding path {fname} to batch image list, please wait..."
                )
                self.log.appendPlainText(
                    f"Adding path {fname} to batch image list"
                )
                qtc.QCoreApplication.processEvents()
                self.imagelistWidget.addItem(extract_fname)
                self.sequential_file_list.append(extract_fname)
                self.sequential_file_dict[extract_fname] = fname
                iid += 1
            else:
                self.statusBar().showMessage(
                    f"Opening file {fname}, please wait..."
                )
                self.log.appendPlainText(f"Opening file {fname}")
                qtc.QCoreApplication.processEvents()
                img = self._open(fname)
                if img is None:
                    continue
                if img.nframes > 1:
                    for img_idx in range(img.nframes):
                        frame = img.getframe(img_idx)
                        self.data_series.append(frame.data[:])
                        self.header_series.append(frame.header.copy())
                        frame_name = f"{extract_fname} # {img_idx}"
                        self.images_list.addItem(frame_name)
                        self.imagelistWidget.addItem(frame_name)
                        self.imgDict[frame_name] = iid
                        self.sequential_file_list.append(frame_name)
                        self.sequential_file_dict[frame_name] = fname
                        iid += 1
                else:
                    self.data_series.append(img.data[:])
                    self.header_series.append(img.header.copy())
                    self.images_list.addItem(extract_fname)
                    self.imagelistWidget.addItem(extract_fname)
                    self.imgDict[extract_fname] = iid
                    self.sequential_file_list.append(extract_fname)
                    self.sequential_file_dict[extract_fname] = fname
                    iid += 1
            self.progressBar.setValue(int((iid + 1) / total * 100.0))
        self.statusBar().clearMessage()
        self.progressBar.setValue(0)
        self.log.appendPlainText("Opening procedure: Complete")
        if self.data_series:
            self.select_new_image(None, imgID=0)
    def open_h5_data_series(self):
        fname = QFileDialog.getOpenFileName(self, "Select and open series of files")
        if isinstance(fname, tuple):
            fname = fname[0]
        fname = str(fname)
        if not fname:
            return
        self.h5_loaded = True
        if self.filecheckBox.checkState():
            self.filecheckBox.stateChanged.disconnect()
            self.filecheckBox.setCheckState(qtc.Qt.Unchecked)
            self.sequential_file_mode = False
            self.filecheckBox.stateChanged.connect(self.sequential_option)
            QMessageBox.warning(
                self,
                "Message",
                "Sequential file mode is not compatible with hdf5 input file: option removed",
            )
        # Reset UI / state
        self.data_series = []
        self.header_series = []
        self.sequential_file_list = []
        self.sequential_file_dict = {}
        self.imagelistWidget.clear()
        self.headerTextEdit.clear()
        with Nexus(fname, "r") as nxs:
            entry = nxs.get_entries()[0]
            nxdata = nxs.get_class(entry, class_type="NXdata")[0]
            dataset = nxdata.get("data", numpy.zeros(shape=(1, 1, 1)))
            total = dataset.shape[0]
            imgDict = {}
            base_name = os.path.basename(os.path.splitext(fname)[0])
            self.images_list.clear()
            safeiid = 0
            for iid in range(total):
                self.progressBar.setValue(int((iid + 1) / total * 100.0))
                self.log.appendPlainText(
                    f"Extracting data from hdf5 archive, image number {iid}"
                )
                qtc.QCoreApplication.processEvents()
                self.data_series.append(dataset[iid])
                self.header_series.append(
                    {"Info": "No header information available in hdf5 Archive"}
                )
                imgDict[f"{base_name}{iid}"] = safeiid
                self.images_list.addItem(f"{base_name}{iid}")
                safeiid += 1
        self.statusBar().clearMessage()
        self.progressBar.setValue(0)
        self.log.appendPlainText("Hdf5 Extraction: Complete")
        self.imgDict = imgDict.copy()
        if self.data_series:
            self.select_new_image(None, imgID=0)
    # ----------------------------------------------------------------------
    # UI helpers
    # ----------------------------------------------------------------------
    def extract_fname_from_path(self, name):
        pos = name.rfind("/")
        return name[pos + 1 :] if pos > -1 else name
    # ----------------------------------------------------------------------
    # Save dialogs & conversion utilities
    # ----------------------------------------------------------------------
    defaultSaveFilter = (
        "binary data block (*.bin);;"
        "cbf image (*.cbf);;"
        "edf image (*.edf);;"
        "oxford diffraction image (*.img);;"
        "mar2300 image(*.mar2300);;"
        "mar3450 image (*.mar3450);;"
        "marccd image (*.marccd));;"
        "tiff image (*.tiff);;"
        "bruker image (*.sfrm)"
    )

    def _getSaveFileNameAndFilter(self, parent=None, caption="", directory="", filter=""):
        dialog = QFileDialog(parent, caption=caption, directory=directory)
        dialog.setAcceptMode(QFileDialog.AcceptSave)
        dialog.setNameFilter(filter)
        if not dialog.exec_():
            return "", ""
        return dialog.selectedFiles()[0], dialog.selectedNameFilter()

    def save_as(self):
        info = self._getSaveFileNameAndFilter(
            self,
            "Save active image as",
            qt.QDir.currentPath(),
            filter=self.tr(self.defaultSaveFilter),
        )
        if self.data.any():
            if info[0] and info[1]:
                fmt = self.extract_format_from_string(str(info[1]))
                fname = self.add_extention_if_absent(str(info[0]), fmt)
                self.convert_and_write(fname, fmt, self.data, self.header)
        else:
            if info[0] and info[1]:
                QMessageBox.warning(
                    self,
                    "Warning",
                    "Could not save image as file if no data have been loaded",
                )
    def save_data_series_as_multiple_file(self):
        info = self._getSaveFileNameAndFilter(
            self,
            "Save data series as multiple files",
            qt.QDir.currentPath(),
            filter=self.tr(self.defaultSaveFilter),
        )
        if self.data_series or self.sequential_file_list:
            if info[0] and info[1]:
                fmt = self.extract_format_from_string(str(info[1]))
                fname = os.path.splitext(str(info[0]))[0]
                self.convert_and_write_multiple_files(fname, fmt)
        else:
            if info[0] and info[1]:
                QMessageBox.warning(
                    self,
                    "Warning",
                    "Could not save image as file if no data have been loaded",
                )
    def save_data_series_as_singlehdf(self):
        info = self._getSaveFileNameAndFilter(
            self,
            "Save data series as single high density file",
            qt.QDir.currentPath(),
            filter=self.tr("HDF5 archive (*.h5)"),
        )
        if self.data_series or self.sequential_file_list:
            if info[0] and info[1]:
                fmt = self.extract_format_from_string(str(info[1]))
                fname = self.add_extention_if_absent(str(info[0]), fmt)
                if fmt == "*.h5":
                    self.convert_and_save_to_h5(fname)
                else:
                    QMessageBox.warning(self, "Warning", f"Unknown format: {fmt}")
        else:
            if info[0] and info[1]:
                QMessageBox.warning(
                    self,
                    "Warning",
                    "Could not save image as file if no data have been loaded",
                )
    # ----------------------------------------------------------------------
    # Conversion / I/O core
    # ----------------------------------------------------------------------
    def convert_and_save_to_h5(self, fname):
        """Save a stack as a Nexus entry (creates a new entry for each image)."""
        with Nexus(fname) as nxs:
            entry = nxs.new_entry(
                entry="entry", program_name="fabio_viewer", title="FabIO Viewer"
            )
            nxdata = nxs.new_class(entry, "fabio", class_type="NXdata")
            # Determine shape & total number of images
            if self.sequential_file_mode:
                total = len(self.sequential_file_list)
                first_fname = self.sequential_file_dict[self.sequential_file_list[0]]
                img = self._open(first_fname)
                if img is None:
                    return
                data = img.data
            else:
                total = len(self.data_series)
                data = self.data_series[0]
            if self.transform_data_series:
                tmpdata = self.apply_queued_transformations(data)
            else:
                tmpdata = data
            shape = tmpdata.shape
            dataset = nxdata.create_dataset(
                "data",
                shape=(total,) + shape,
                dtype=numpy.float32,
                chunks=(1,) + shape,
                compression="gzip",
            )
            dataset.attrs["interpretation"] = "image"
            dataset.attrs["signal"] = "1"
            if self.sequential_file_mode:
                for iid, imgkey in enumerate(self.sequential_file_list):
                    tmpfname = self.sequential_file_dict[imgkey]
                    img = self._open(tmpfname)
                    if img is None:
                        continue
                    self.progressBar.setValue(int((iid + 1) / total * 100.0))
                    self.log.appendPlainText(
                        f"Converting and saving file {tmpfname}. saving file number {iid}"
                    )
                    qtc.QCoreApplication.processEvents()
                    if self.transform_data_series:
                        tmpdata = self.apply_queued_transformations(img.data)
                    else:
                        tmpdata = img.data
                    dataset[iid] = tmpdata
            else:
                for iid, data in enumerate(self.data_series):
                    self.log.appendPlainText(f"Saving file number {iid}")
                    self.progressBar.setValue(int((iid + 1) / total * 100.0))
                    qtc.QCoreApplication.processEvents()
                    if self.transform_data_series:
                        tmpdata = self.apply_queued_transformations(data)
                    else:
                        tmpdata = data
                    dataset[iid] = tmpdata
        self.statusBar().clear()
        self.progressBar.setValue(0)
        self.log.appendPlainText("Hdf5 Recording: Complete")
    def add_extention_if_absent(self, fname, fmt):
        """Append the required extension if the filename does not already have one."""
        slash = fname.rfind("/")
        dot = fname.rfind(".")
        return fname if dot > slash else fname + fmt[1:]
    def convert_and_write(self, fname, fmt, data, header):
        """Write a single image in the requested format."""
        if fmt == "*.bin":
            with open(fname, mode="wb") as out:
                data.tofile(out)
            return
        elif fmt == "*.marccd":
            out = fabio.marccdimage.marccdimage(data=data, header=header)
        elif fmt == "*.edf":
            out = fabio.edfimage.edfimage(data=data, header=header)
        elif fmt == "*.tiff":
            out = fabio.tifimage.tifimage(data=data, header=header)
        elif fmt == "*.cbf":
            out = fabio.cbfimage.cbfimage(data=data, header=header)
        elif fmt in ("*.mar3450", "*.mar2300"):
            data = self.padd_mar(data, fmt)
            out = fabio.mar345image.mar345image(data=data, header=header)
        elif fmt == "*.img":
            out = fabio.OXDimage.OXDimage(data=data, header=header)
        elif fmt == "*.sfrm":
            out = fabio.brukerimage.brukerimage(data=data, header=header)
        else:
            raise Warning(f"Unknown format: {fmt}")
        self.statusBar().showMessage(
            f"Writing file {fname} to {fmt[2:]} format, please wait..."
        )
        self.log.appendPlainText(f"Writing file {fname} to {fmt[2:]} format")
        qtc.QCoreApplication.processEvents()
        out.write(fname)
        self.statusBar().clearMessage()

    def convert_and_write_multiple_files(self, fname, fmt):
        """Write a series of images to separate files."""
        if self.sequential_file_mode:
            total = len(self.sequential_file_list)
            ii = 0
            for imgkey in self.sequential_file_list:
                tmpfname = self.sequential_file_dict[imgkey]
                img = self._open(tmpfname)
                if img is None:
                    continue
                self.progressBar.setValue(int((ii + 1) / total * 100.0))
                self.log.appendPlainText(f"Converting file {tmpfname}")
                qtc.QCoreApplication.processEvents()
                if self.transform_data_series:
                    tmpdata = self.apply_queued_transformations(img.data)
                else:
                    tmpdata = img.data
                outname = (f"{fname}_{self.counter_format}{fmt[1:]}") % ii
                self.convert_and_write(outname, fmt, tmpdata, img.header)
                ii += 1
        else:
            total = len(self.data_series)
            for i in range(total):
                tmpdata = self.data_series[i]
                tmpheader = self.header_series[i]
                outname = (f"{fname}_{self.counter_format}{fmt[1:]}") % i
                self.progressBar.setValue(int((i + 1) / total * 100.0))
                self.log.appendPlainText(f"Converting file {i}")
                qtc.QCoreApplication.processEvents()
                if self.transform_data_series:
                    tmpdata = self.apply_queued_transformations(tmpdata)
                self.convert_and_write(outname, fmt, tmpdata, tmpheader)
        self.progressBar.setValue(0)
        self.log.appendPlainText(f"Convertion to {fmt[2:]}: Complete")
    def extract_format_from_string(self, fmt_long):
        for fmt in output_format:
            if fmt in fmt_long:
                return fmt
        raise Warning(f"Unknown format: {fmt_long}")
    # ----------------------------------------------------------------------
    # Image transformations
    # ----------------------------------------------------------------------
    def horizontal_mirror(self):
        if self.transform_data_series:
            if self.sequential_file_mode:
                self.transformation_queue.addItem("horizontal_mirror")
                self.transform_list.append("horizontal_mirror")
                self.log.appendPlainText(
                    "Add horizontal mirror to transformations queue"
                )
                qtc.QCoreApplication.processEvents()
            else:
                total = len(self.data_series)
                if not total:
                    QMessageBox.warning(
                        self,
                        "Warning",
                        "Could not transform image if no data have been loaded",
                    )
                    return
                for i in range(total):
                    self.data_series[i] = numpy.flipud(self.data_series[i])[:]
                    self.progressBar.setValue(int((i + 1) / total * 100.0))
                    self.log.appendPlainText(
                        f"Applying horizontal mirror to data series: image {i}"
                    )
                    qtc.QCoreApplication.processEvents()
                iid = self.imgDict[str(self.images_list.currentText())]
                self.select_new_image(None, imgID=iid)
        else:
            if self.data.any():
                self.data = numpy.flipud(self.data)[:]
                iid = self.imgDict[str(self.images_list.currentText())]
                self.data_series[iid] = self.data[:]
                self.log.appendPlainText("Applying horizontal mirror to current data")
                self.on_draw()
            else:
                QMessageBox.warning(
                    self,
                    "Warning",
                    "Could not transform image if no data have been loaded",
                )
        self.progressBar.setValue(0)

    def vertical_mirror(self):
        if self.transform_data_series:
            if self.sequential_file_mode:
                self.transformation_queue.addItem("vertical_mirror")
                self.transform_list.append("vertical_mirror")
                self.log.appendPlainText(
                    "Add vertical mirror to transformations queue"
                )
                qtc.QCoreApplication.processEvents()
            else:
                total = len(self.data_series)
                if not total:
                    QMessageBox.warning(
                        self,
                        "Warning",
                        "Could not transform image if no data have been loaded",
                    )
                    return
                for i in range(total):
                    self.data_series[i] = numpy.fliplr(self.data_series[i])[:]
                    self.progressBar.setValue(int((i + 1) / total * 100.0))
                    self.log.appendPlainText(
                        f"Applying vertical mirror to data series: image {i}"
                    )
                    qtc.QCoreApplication.processEvents()
                iid = self.imgDict[str(self.images_list.currentText())]
                self.select_new_image(None, imgID=iid)
        else:
            if self.data.any():
                self.data = numpy.fliplr(self.data)[:]
                iid = self.imgDict[str(self.images_list.currentText())]
                self.data_series[iid] = self.data[:]
                self.log.appendPlainText("Applying vertical mirror to current data")
                self.on_draw()
            else:
                QMessageBox.warning(
                    self,
                    "Warning",
                    "Could not transform image if no data have been loaded",
                )
        self.progressBar.setValue(0)

    def transposition(self):
        if self.transform_data_series:
            if self.sequential_file_mode:
                self.transformation_queue.addItem("transposition")
                self.transform_list.append("transposition")
                self.log.appendPlainText(
                    "Add transposition to transformations queue"
                )
                qtc.QCoreApplication.processEvents()
            else:
                total = len(self.data_series)
                if not total:
                    QMessageBox.warning(
                        self,
                        "Warning",
                        "Could not transform image if no data have been loaded",
                    )
                    return
                for i in range(total):
                    self.data_series[i] = self.data_series[i].transpose()[:]
                    self.progressBar.setValue(int((i + 1) / total * 100.0))
                    self.log.appendPlainText(
                        f"Applying transposition to data series: image {i}"
                    )
                    qtc.QCoreApplication.processEvents()
                iid = self.imgDict[str(self.images_list.currentText())]
                self.select_new_image(None, imgID=iid)
        else:
            if self.data.any():
                self.data = self.data.transpose()[:]
                iid = self.imgDict[str(self.images_list.currentText())]
                self.data_series[iid] = self.data[:]
                self.log.appendPlainText("Applying transposition to current data")
                self.on_draw()
            else:
                QMessageBox.warning(
                    self,
                    "Warning",
                    "Could not transform image if no data have been loaded",
                )
        self.progressBar.setValue(0)

    def rotation_90(self):
        if self.transform_data_series:
            if self.sequential_file_mode:
                self.transformation_queue.addItem("rotation(+90)")
                self.transform_list.append("rotation(+90)")
                self.log.appendPlainText("Add + 90 rotation to transformations queue")
                qtc.QCoreApplication.processEvents()
            else:
                total = len(self.data_series)
                if not total:
                    QMessageBox.warning(
                        self,
                        "Warning",
                        "Could not transform image if no data have been loaded",
                    )
                    return
                for i in range(total):
                    self.data_series[i] = numpy.rot90(self.data_series[i])[:]
                    self.progressBar.setValue(int((i + 1) / total * 100.0))
                    self.log.appendPlainText(
                        f"Applying + 90 rotation to data series: image {i}"
                    )
                    qtc.QCoreApplication.processEvents()
                iid = self.imgDict[str(self.images_list.currentText())]
                self.select_new_image(None, imgID=iid)
        else:
            if self.data.any():
                self.data = numpy.rot90(self.data)[:]
                iid = self.imgDict[str(self.images_list.currentText())]
                self.data_series[iid] = self.data[:]
                self.log.appendPlainText("Applying + 90 rotation to current data")
                self.on_draw()
            else:
                QMessageBox.warning(
                    self,
                    "Warning",
                    "Could not transform image if no data have been loaded",
                )
        self.progressBar.setValue(0)

    def rotation_180(self):
        if self.transform_data_series:
            if self.sequential_file_mode:
                self.transformation_queue.addItem("rotation(+180)")
                self.transform_list.append("rotation(+180)")
                self.log.appendPlainText("Add + 180 rotation to transformations queue")
                qtc.QCoreApplication.processEvents()
            else:
                total = len(self.data_series)
                if not total:
                    QMessageBox.warning(
                        self,
                        "Warning",
                        "Could not transform image if no data have been loaded",
                    )
                    return
                for i in range(total):
                    self.data_series[i] = numpy.rot90(self.data_series[i], 2)[:]
                    self.progressBar.setValue(int((i + 1) / total * 100.0))
                    self.log.appendPlainText(
                        f"Applying + 180 rotation to data series: image {i}"
                    )
                    qtc.QCoreApplication.processEvents()
                iid = self.imgDict[str(self.images_list.currentText())]
                self.select_new_image(None, imgID=iid)
        else:
            if self.data.any():
                self.data = numpy.rot90(self.data, 2)[:]
                iid = self.imgDict[str(self.images_list.currentText())]
                self.data_series[iid] = self.data[:]
                self.log.appendPlainText("Applying + 180 rotation to current data")
                self.on_draw()
            else:
                QMessageBox.warning(
                    self,
                    "Warning",
                    "Could not transform image if no data have been loaded",
                )
        self.progressBar.setValue(0)

    def rotation_270(self):
        if self.transform_data_series:
            if self.sequential_file_mode:
                self.transformation_queue.addItem("rotation(-90)")
                self.transform_list.append("rotation(-90)")
                self.log.appendPlainText("Add - 90 rotation to transformations queue")
                qtc.QCoreApplication.processEvents()
            else:
                total = len(self.data_series)
                if not total:
                    QMessageBox.warning(
                        self,
                        "Warning",
                        "Could not transform image if no data have been loaded",
                    )
                    return
                for i in range(total):
                    self.data_series[i] = numpy.rot90(self.data_series[i], 3)[:]
                    self.progressBar.setValue(int((i + 1) / total * 100.0))
                    self.log.appendPlainText(
                        f"Applying - 90 rotation to data series: image {i}"
                    )
                    qtc.QCoreApplication.processEvents()
                iid = self.imgDict[str(self.images_list.currentText())]
                self.select_new_image(None, imgID=iid)
        else:
            if self.data.any():
                self.data = numpy.rot90(self.data, 3)[:]
                iid = self.imgDict[str(self.images_list.currentText())]
                self.data_series[iid] = self.data[:]
                self.log.appendPlainText("Applying - 90 rotation to current data")
                self.on_draw()
            else:
                QMessageBox.warning(
                    self,
                    "Warning",
                    "Could not transform image if no data have been loaded",
                )
        self.progressBar.setValue(0)

    def mask(self):
        fname = QFileDialog.getOpenFileName(self, "Select and import a boolean mask from binary data block file")
        if isinstance(fname, tuple):
            fname = fname[0]

        fname = str(fname)
        if not fname:
            return
        dim1, dim2, offset, bytecode, endian = BinDialog(self).exec_()
        if dim1 is None:
            return
        endian = "<" if endian == "Short" else ">"
        img = fabio.binaryimage.binaryimage()
        img.read(fname, dim1, dim2, offset, bytecode, endian)
        self.mask = img.data[:]
        if self.transform_data_series:
            if self.sequential_file_mode:
                self.transformation_queue.addItem("masking")
                self.transform_list.append("masking")
                self.log.appendPlainText("Add masking to transformations queue")
                qtc.QCoreApplication.processEvents()
            else:
                total = len(self.data_series)
                if not total:
                    QMessageBox.warning(
                        self,
                        "Warning",
                        "Could not transform image if no data have been loaded",
                    )
                    return
                for i in range(total):
                    if self.data_series[i].shape != self.mask.shape:
                        msg = f"Mask and image have different shapes, skipping image {i}"
                        QMessageBox.warning(self, "Warning", msg)
                        self.log.appendPlainText(msg)
                    else:
                        self.data_series[i] = self.mask * self.data_series[i]
                        self.progressBar.setValue(int((i + 1) / total * 100.0))
                        self.log.appendPlainText(
                            f"Applying mask to data series: image {i}"
                        )
                        qtc.QCoreApplication.processEvents()
                iid = self.imgDict[str(self.images_list.currentText())]
                self.select_new_image(None, imgID=iid)
        else:
            if self.data.any():
                self.data = self.mask * self.data
                iid = self.imgDict[str(self.images_list.currentText())]
                self.data_series[iid] = self.data[:]
                self.on_draw()
                self.statusBar().showMessage("Binary boolean mask loaded and applied", 2000)
                self.log.appendPlainText("Binary boolean mask loaded and applied")
            else:
                QMessageBox.warning(
                    self,
                    "Warning",
                    "Could not transform image if no data have been loaded",
                )
        self.progressBar.setValue(0)
    def downsample(self):
        thick, start_angle, step_angle = DownSamplingDialog().exec_()
        if thick is None:
            return
        info = self._getSaveFileNameAndFilter(
            self,
            "Save downsampled data series as multiple files",
            qt.QDir.currentPath(),
            filter=self.tr(self.defaultSaveFilter),
        )
        if not (self.data_series or self.sequential_file_list):
            if info[0] and info[1]:
                QMessageBox.warning(
                    self,
                    "Warning",
                    "Could not save image as file if no data have been loaded",
                )
            return
        if not (info[0] and info[1]):
            return
        fmt = self.extract_format_from_string(str(info[1]))
        fname = os.path.splitext(str(info[0]))[0]
        if self.sequential_file_mode:
            total = len(self.sequential_file_list)
            stack = None
            subtotal = (total // thick) * thick
            for i in range(subtotal):
                j = i % thick
                k = i // thick
                imgkey = self.sequential_file_list[i]
                tmpfname = self.sequential_file_dict[imgkey]
                img = self._open(tmpfname)
                if img is None:
                    continue
                if stack is None:
                    stack = numpy.zeros_like(img.data)
                if img.data.shape != stack.shape:
                    self.log.appendPlainText(
                        f"Error image shape: {img.data.shape} summed data shape: {stack.shape}"
                    )
                    continue
                numpy.add(stack, img.data, stack)
                self.progressBar.setValue(int((i + 1) / subtotal * 100.0))
                self.log.appendPlainText(f"File {imgkey} stacked")
                qtc.QCoreApplication.processEvents()
                if j == thick - 1:
                    self.log.appendPlainText(f"stack number {k} summing up")
                    qtc.QCoreApplication.processEvents()
                    if fmt in ("*.mar3450", "*.mar2300"):
                        img.header["PHI_START"] = f"{start_angle + step_angle * (i - thick + 1):.3f}"
                        img.header["PHI_END"] = f"{start_angle + step_angle * i:.3f}"
                    outname = (f"{fname}_{self.counter_format}{fmt[1:]}") % k
                    self.convert_and_write(outname, fmt, stack, img.header)
                    stack = numpy.zeros_like(img.data)
        else:
            total = len(self.data_series)
            stack = numpy.zeros_like(self.data_series[0])
            subtotal = (total // thick) * thick
            for i in range(subtotal):
                j = i % thick
                k = i // thick
                data = self.data_series[i]
                if data.shape != stack.shape:
                    self.log.appendPlainText(
                        f"Error image shape: {data.shape} summed data shape: {stack.shape}"
                    )
                    continue
                numpy.add(stack, data, stack)
                self.progressBar.setValue(int((i + 1) / subtotal * 100.0))
                self.log.appendPlainText(f"File number {i} stacked")
                qtc.QCoreApplication.processEvents()
                if j == thick - 1:
                    self.log.appendPlainText(f"stack number {k} summing up")
                    qtc.QCoreApplication.processEvents()
                    if fmt in ("*.mar3450", "*.mar2300"):
                        self.header_series[i]["PHI_START"] = f"{start_angle + step_angle * (i - thick + 1):.3f}"
                        self.header_series[i]["PHI_END"] = f"{start_angle + step_angle * i:.3f}"
                    outname = (f"{fname}_{self.counter_format}{fmt[1:]}") % k
                    self.convert_and_write(outname, fmt, stack, self.header_series[i])
                    stack = numpy.zeros_like(data)
        self.progressBar.setValue(0)
        self.log.appendPlainText("Downsampling: Complete")
        qtc.QCoreApplication.processEvents()
    # ----------------------------------------------------------------------
    # UI callbacks – image selection, drawing, etc.
    # ----------------------------------------------------------------------
    def select_new_image(self, name, imgID=None):
        iid = imgID if imgID is not None else self.imgDict[str(name)]
        self.data = self.data_series[iid]
        self.header = self.header_series[iid]

        self.headerTextEdit.setPlainText(str(self.format_header(self.header)))
        self.on_draw()

    def on_pick(self, event):
        if event.inaxes and self.data.any():
            x = int(round(event.xdata))
            y = int(round(event.ydata))
            if x < self.data.shape[1] and y < self.data.shape[0]:
                i = self.data[y, x]
                self.pix_coords_label.setText(
                    f"Pixel coordinates and intensity: x ={x:6d}, y ={y:6d}, i ={i:6g}"
                )
            else:
                self.pix_coords_label.setText(
                    "Pixel coordinates and intensity: x = None , y = None , i = None "
                )
        else:
            self.pix_coords_label.setText(
                "Pixel coordinates and intensity: x = None , y = None , i = None "
            )

    def on_draw(self):
        """Redraws the figure"""
        self.statusBar().showMessage("Loading display...")
        qtc.QCoreApplication.processEvents()
        self.axes.clear()
        self.axes.imshow(numpy.arcsinh(self.data), interpolation="nearest")

        self.axes.set_visible(True)
        if self.axes.get_ylim()[0] < self.axes.get_ylim()[1]:
            self.axes.set_ylim(self.axes.get_ylim()[::-1])
        self.canvas.draw()
        self.statusBar().clearMessage()

    def batch_to_view(self):
        items = self.imagelistWidget.selectedItems()
        if not items:
            return
        item = str(items[0].text())
        if any(ext in item.lower() for ext in (".h5", ".hdf", ".hdf5")):
            QMessageBox.warning(
                self, "Message", "Can't display hdf archive from batch mode "
            )
            return
        if self.sequential_file_mode:
            self.data_series = []
            self.header_series = []
            self.imgDict = {}
            self.images_list.clear()
            self.headerTextEdit.clear()
            self.axes.clear()
            self.canvas.draw()
            self.statusBar().showMessage(
                f"Import image {item} in the View Mode tab, please wait..."
            )
            self.log.appendPlainText(f"Import image {item} in the View Mode tab")
            qtc.QCoreApplication.processEvents()
            fname = self.sequential_file_dict[item]
            extract_fname = os.path.splitext(os.path.basename(fname))[0]
            img = self._open(fname)
            if img is None:
                return

            if img.nframes > 1:
                for img_idx in range(img.nframes):
                    frame = img.getframe(img_idx)
                    self.data_series.append(frame.data[:])
                    self.header_series.append(frame.header.copy())
                    frame_name = f"{extract_fname} # {img_idx}"
                    self.images_list.addItem(frame_name)
                    self.imagelistWidget.addItem(frame_name)
                    self.imgDict[frame_name] = len(self.data_series) - 1
            else:
                self.data_series.append(img.data[:])
                self.header_series.append(img.header.copy())
                self.images_list.addItem(extract_fname)
                self.imagelistWidget.addItem(extract_fname)
                self.imgDict[extract_fname] = len(self.data_series) - 1
        self.statusBar().clearMessage()
        if self.data_series:
            self.select_new_image(None, imgID=0)
        self.tabWidget.setCurrentIndex(0)

    def set_counter_format_option(self):
        self.counter_format = CounterFormatOptionDialog(self.counter_format).exec_()
    def padd_mar(self, data, fmt):
        """Pad or crop MAR images to the required 2300×2300 or 3450×3450 size."""
        dim1, dim2 = data.shape
        size = 2300 if fmt == "*.mar2300" else 3450
        left = (size - dim1) // 2
        right = size - (dim1 + left)
        up = (size - dim2) // 2
        down = size - (dim2 + up)

        out = numpy.zeros((size, size))
        # Compute slice indices (positive → pad, negative → crop)
        in_lm, out_lm = (0, left) if left > 0 else (-left, 0)
        in_rm, out_rm = (dim1, size) if right > 0 else (dim1 - right, size)
        in_um, out_um = (0, up) if up > 0 else (-up, 0)
        in_dm, out_dm = (dim2, size) if down > 0 else (dim2 - down, size)
        out[out_lm:out_rm, out_um:out_dm] = data[in_lm:in_rm, in_um:in_dm]
        return out

    def sequential_option(self, state):
        if not self.h5_loaded:
            self.sequential_file_mode = state == qtc.Qt.Checked
        else:
            self.filecheckBox.stateChanged.disconnect()
            self.filecheckBox.setCheckState(qtc.Qt.Unchecked)
            self.sequential_file_mode = False
            self.filecheckBox.stateChanged.connect(self.sequential_option)
            QMessageBox.warning(
                self,
                "Message",
                "Sequential file mode is not compatible with hdf5 input file: option removed",
            )
    # ----------------------------------------------------------------------
    # Menu / toolbar creation
    # ----------------------------------------------------------------------
    def create_menu(self):
        self.file_menu = self.menuBar().addMenu("&File")
        self.open_menu = self.file_menu.addMenu("&Open")
        # ---- Open ----
        self.add_actions(
            self.open_menu,
            (
                self.create_action(
                    "&Image(s)",
                    slot=self.open_data_series,
                    tip="Load single file and data series (files sequence)",
                ),
                self.create_action(
                    "&Hdf5 data series",
                    slot=self.open_h5_data_series,
                    tip="Load a series stored in an HDF5 file",
                ),
            ),
        )
        # ---- Save ----
        self.save_as_menu = self.file_menu.addMenu("&Save")
        self.add_actions(
            self.save_as_menu,
            (
                self.create_action(
                    "&Active image",
                    slot=self.save_as,
                    tip="Save/Convert the image which is currently displayed",
                ),
            ),
        )
        self.save_data_series_menu = self.save_as_menu.addMenu("&Data series as")
        self.add_actions(
            self.save_data_series_menu,
            (
                self.create_action(
                    "&Multiple files",
                    slot=self.save_data_series_as_multiple_file,
                    tip="Save/Convert the set of images currently loaded into the images list",
                ),
                self.create_action(
                    "&Hdf5 archive",
                    slot=self.save_data_series_as_singlehdf,
                    tip="Save/Convert the set of images currently loaded into the images list",
                ),
            ),
        )
        self.add_actions(
            self.file_menu,
            (
                self.create_action(
                    "&Quit", slot=self.close, shortcut="Ctrl+Q", tip="Close the application"
                ),
            ),
        )
        # ---- Transform ----
        self.transform_menu = self.menuBar().addMenu("&Transform")
        self.mirror_menu = self.transform_menu.addMenu("&Mirror")
        self.add_actions(
            self.mirror_menu,
            (
                self.create_action(
                    "&Horizontal", slot=self.horizontal_mirror, tip="Horizontal mirror"
                ),
                self.create_action(
                    "&Vertical", slot=self.vertical_mirror, tip="Vertical mirror"
                ),
                self.create_action(
                    "&Transposition", slot=self.transposition, tip="Transposition"
                ),
            ),
        )
        self.rotation_menu = self.transform_menu.addMenu("&Rotation")
        self.add_actions(
            self.rotation_menu,
            (
                self.create_action(
                    "+90", slot=self.rotation_90, tip="Rotation of +90 degrees (counter‑clockwise)"
                ),
                self.create_action(
                    "+180", slot=self.rotation_180, tip="Rotation of +180 degrees (counter‑clockwise)"
                ),
                self.create_action(
                    "- 90", slot=self.rotation_270, tip="Rotation of -90 degrees (counter‑clockwise)"
                ),
            ),
        )
        self.add_actions(
            self.transform_menu,
            (
                self.create_action(
                    "&Mask", slot=self.mask, tip="Import a mask from file and apply it to image(s)"
                ),
                self.create_action(
                    "&Downsample",
                    slot=self.downsample,
                    tip="Summation over groups of images",
                ),
                self.create_action(
                    "&Apply transform to the whole data series",
                    slot=self.transformation_options,
                    tip=(
                        "If checked, transformations are applied to the whole data series; "
                        "if unchecked, only the active image is affected"
                    ),
                ),
            ),
        )
        self.transform_option_action = self.transform_menu.actions()[-1]
        self.transform_option_action.setCheckable(True)
        # ---- Options ----
        self.options = self.menuBar().addMenu("&Options")
        self.add_actions(
            self.options,
            (
                self.create_action(
                    "&Counter format",
                    slot=self.set_counter_format_option,
                    tip="Define the format for the counter in multiple saving",
                ),
            ),
        )
        # ---- Help ----
        self.help_menu = self.menuBar().addMenu("&Help")
        self.add_actions(
            self.help_menu,
            (
                self.create_action(
                    "&About", slot=self.on_about, shortcut="F1", tip="About FabIO Viewer"
                ),
            ),
        )
    def add_actions(self, target, actions):
        for action in actions:
            if action is None:
                target.addSeparator()
            else:
                target.addAction(action)
    def create_action(
        self,
        text,
        slot=None,
        shortcut=None,
        icon=None,
        tip=None,
        checkable=False,
        signal="triggered",
    ):
        action = QAction(text, self)
        if icon is not None:
            action.setIcon(QIcon(icon))
        if shortcut is not None:
            action.setShortcut(shortcut)
        if tip is not None:
            action.setToolTip(tip)
            action.setStatusTip(tip)
        if slot is not None:
            getattr(action, signal).connect(slot)
        action.setCheckable(checkable)
        return action
    # ----------------------------------------------------------------------
    # Main UI layout (tabs, canvases, log, etc.)
    # ----------------------------------------------------------------------
    def create_main_frame(self):
        self.tabWidget = QTabWidget()
        tab1 = QWidget()
        self.tabWidget.addTab(tab1, "View Mode")
        tab2 = QWidget()
        self.tabWidget.addTab(tab2, "Batch Mode")
        # ------------------------------------------------------------------
        # Matplotlib canvas (View Mode)
        # ------------------------------------------------------------------
        self.dpi = 100
        self.fig = Figure(dpi=self.dpi)
        self.canvas = FigureCanvasQTAgg(self.fig)
        self.canvas.setParent(tab1)

        # Since we have only one plot, we can use add_axes
        # instead of add_subplot, but then the subplot
        # configuration tool in the navigation toolbar wouldn't
        # work.
        #
        self.axes = self.fig.add_subplot(111)
        self.axes.set_visible(False)
        # Bind the 'pick' event for clicking on one of the bars
        self.canvas.mpl_connect("motion_notify_event", self.on_pick)

        # Create the navigation toolbar, tied to the canvas
        self.mpl_toolbar = NavigationToolbar2QT(self.canvas, tab1, coordinates=False)
        # ------------------------------------------------------------------
        # Header display (left side)
        # ------------------------------------------------------------------
        header_vbox = QVBoxLayout()
        header_label = QLabel("Header Info:", self)
        header_vbox.addWidget(header_label)
        self.headerTextEdit = QPlainTextEdit(tab1)
        self.headerTextEdit.setReadOnly(True)
        header_vbox.addWidget(self.headerTextEdit)
        # ------------------------------------------------------------------
        # Image selector & viewer (right side)
        # ------------------------------------------------------------------
        selector_label = QLabel("Active Image:", self)
        self.images_list = QComboBox(self)
        self.images_list.textActivated.connect(self.select_new_image)
        viewer_label = QLabel("Images Viewer:", self)
        self.pix_coords_label = QLabel(
            "Pixel coordinates and intensity: x = None , y = None , i = None ", self
        )
        self.mpl_toolbar.addWidget(self.pix_coords_label)
        hbox = QHBoxLayout()
        hbox.addWidget(selector_label, alignment=qtc.Qt.AlignRight)
        hbox.addWidget(self.images_list)
        vbox = QVBoxLayout()
        vbox.addWidget(viewer_label, alignment=qtc.Qt.AlignVCenter)
        vbox.addWidget(self.canvas, alignment=qtc.Qt.AlignVCenter)
        vbox.addWidget(self.mpl_toolbar, alignment=qtc.Qt.AlignVCenter)
        vbox.addLayout(hbox)
        left = QWidget()
        right = QWidget()
        left.setLayout(header_vbox)
        right.setLayout(vbox)
        splitter = QSplitter(qtc.Qt.Horizontal)
        splitter.addWidget(left)
        splitter.addWidget(right)
        main_hbox = QHBoxLayout()
        main_hbox.addWidget(splitter)
        tab1.setLayout(main_hbox)
        # ------------------------------------------------------------------
        # Batch Mode UI (tab 2)
        # ------------------------------------------------------------------
        imagelistvbox = QVBoxLayout()
        imagelistlabel = QLabel("Images List:", tab2)
        self.imagelistWidget = QListWidget(tab2)
        import_view_button = QPushButton("Export image to View Mode", tab2)
        import_view_button.clicked.connect(self.batch_to_view)

        imagelistvbox.addWidget(imagelistlabel)
        imagelistvbox.addWidget(self.imagelistWidget)
        imagelistvbox.addWidget(import_view_button)
        # File‑mode options
        optiongroupBox = QGroupBox("File Modes:", tab2)
        optionbox = QVBoxLayout()
        self.butttonGroup = QButtonGroup()
        self.filecheckBox = QCheckBox("Sequential access (for large data series)", tab2)
        self.filecheckBox.stateChanged.connect(self.sequential_option)
        self.butttonGroup.addButton(self.filecheckBox)
        self.filecheckBox2 = QCheckBox(
            "Direct access (all images are stored in memory simultaneously)", tab2
        )
        self.filecheckBox2.setChecked(True)
        self.butttonGroup.addButton(self.filecheckBox2)

        self.butttonGroup.setExclusive(True)

        optionbox.addWidget(self.filecheckBox)
        optionbox.addWidget(self.filecheckBox2)

        optiongroupBox.setLayout(optionbox)
        # Transformations queue
        queuegroupBox = QGroupBox("Transformations Queue:", tab2)
        queuebox = QVBoxLayout()
        self.transformation_queue = QListWidget(tab2)
        queuebox.addWidget(self.transformation_queue)
        clear_trans_list_button = QPushButton("Clear Transformation List", tab2)
        clear_trans_list_button.clicked.connect(self.clear_transform_list)
        queuebox.addWidget(clear_trans_list_button)

        queuegroupBox.setLayout(queuebox)
        # Log view
        loggroupBox = QGroupBox("Log View:", tab2)
        logbox = QHBoxLayout()
        self.log = QPlainTextEdit()
        logbox.addWidget(self.log)

        loggroupBox.setLayout(logbox)
        # Progress bar
        self.progressBar = QProgressBar(tab2)
        self.progressBar.setProperty("value", 0)
        # Assemble right‑hand side of Batch tab
        rightsidevbox = QVBoxLayout()
        rightsidevbox.addWidget(optiongroupBox)
        rightsidevbox.addWidget(queuegroupBox)
        rightsidevbox.addWidget(loggroupBox)
        rightsidevbox.addWidget(self.progressBar)
        left2 = QWidget()
        right2 = QWidget()
        left2.setLayout(imagelistvbox)
        right2.setLayout(rightsidevbox)
        splitter2 = QSplitter(qtc.Qt.Horizontal)
        splitter2.addWidget(left2)
        splitter2.addWidget(right2)
        splitter2.setStretchFactor(1, 2)
        batch_hbox = QHBoxLayout()
        batch_hbox.addWidget(splitter2)
        tab2.setLayout(batch_hbox)
        self.setCentralWidget(self.tabWidget)

    def create_status_bar(self):
        self.status_text = QLabel("")
        self.statusBar().addWidget(self.status_text, 1)
        self.statusBar().showMessage("Thanks for using FabIO viewer.", 5000)

    def on_about(self):
        msg = [
            __doc__,
            "",
            f"Version: \t\t{__version__}",
            f"FabIO version: \t{fabio.version}",
            f"Author: \t\t{__author__}",
            f"Copyright: \t\t{__copyright__}",
            f"License: \t\t{__licence__}",
        ]
        QMessageBox.about(self, "About FabIO Viewer", os.linesep.join(msg))
    # ----------------------------------------------------------------------
    # Miscellaneous UI helpers
    # ----------------------------------------------------------------------
    def clear_transform_list(self):
        self.transform_list = []
        self.transformation_queue.clear()
    def transformation_options(self):
        self.transform_data_series = self.transform_option_action.isChecked()



#!/usr/bin/env python

"""
Test import all submodules
"""

import os
import unittest
import logging

logger = logging.getLogger(__name__)


class TestImport(unittest.TestCase):
    def test_import_all(self):
        import fabio

        base = os.path.split(fabio.__path__[0])[0] + "/"
        for root, dirs, files in os.walk(fabio.__path__[0]):
            for f in files:
                if f.endswith(".py"):
                    module = os.path.join(root, f[:-3])[len(base) :].replace(
                        os.sep, "."
                    )
                    __import__(module)


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestImport))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.#

"""Test lambda images"""

import os
import numpy
import fabio.lambdaimage
from fabio.openimage import openimage
from ..utilstest import UtilsTest
import unittest
import logging

logger = logging.getLogger(__name__)


class TestLambda(unittest.TestCase):
    # filename dim1 dim2 min max mean stddev
    TESTIMAGES = [
        ("l1_test02_00002_m01.nxs", 1554, 516, 0, 548, 0.00, 0.81024),  # WIP
        ("l1_test02_00002_m02.nxs", 1554, 516, 0, 0, 0.0, 0.0),  # WIP
        ("l1_test02_00002_m03.nxs", 1554, 516, 0, 45, 0.00, 0.0534),  # WIP
        ("l1_test02_00002_master.nxs", 1555, 1813, 0, 548, 0.00, 0.433),  # WIP
    ]

    def test_read(self):
        """
        Test the reading of lambda images
        """
        for params in self.TESTIMAGES:
            name = params[0]
            logger.debug("Processing: %s" % name)
            dim1, dim2 = params[1:3]
            shape = dim2, dim1
            mini, maxi, mean, stddev = params[3:]
            obj = fabio.lambdaimage.LambdaImage()
            filename = UtilsTest.getimage(name)
            obj.read(filename)

            self.assertAlmostEqual(
                mini, obj.getmin(), 2, "getmin [%s,%s]" % (mini, obj.getmin())
            )
            self.assertAlmostEqual(
                maxi, obj.getmax(), 2, "getmax [%s,%s]" % (maxi, obj.getmax())
            )
            self.assertAlmostEqual(
                mean, obj.getmean(), 2, "getmean [%s,%s]" % (mean, obj.getmean())
            )
            self.assertAlmostEqual(
                stddev,
                obj.getstddev(),
                2,
                "getstddev [%s,%s]" % (stddev, obj.getstddev()),
            )

            self.assertEqual(shape, obj.shape, "dim1")

    def test_write(self):
        """read file using fabio.open and save and reopen ... check consistency"""
        for params in self.TESTIMAGES:
            fname = UtilsTest.getimage(params[0])
            obj = openimage(fname)
            self.assertEqual(obj.shape, params[2:0:-1])

            dst = os.path.join(UtilsTest.tempdir, os.path.basename(fname))
            obj.write(dst)
            for idx, read_back in enumerate(openimage(dst)):
                self.assertTrue(
                    numpy.all(read_back.data == obj.getframe(idx).data),
                    f"data are the same {fname} #{idx}",
                )


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestLambda))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/python
# coding: utf-8
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Benchmark for file reading"""

__author__ = "Jérôme Kieffer"
__date__ = "27/10/2025"
__license__ = "MIT"
__copyright__ = "2016-2020 European Synchrotron Radiation Facility, Grenoble, France"

import sys
import timeit
import os
from ..test import utilstest

# To use use the locally build version of PyFAI, use ../bootstrap.py
try:
    from .. import open as fabio_open, version, date
except ImportError:
    from fabio import open as fabio_open, version, date

datasets = [
    "mb_LP_1_001.img",
    "Cr8F8140k103.0026",
    "run2_1_00148.cbf",
    "F2K_Seb_Lyso0675.edf",
    "fit2d_click.msk",
    "GE_aSI_detector_image_1529",
    "i01f0001.kcd",
    "example.mar2300",
    "corkcont2_H_0089.mccd",
    "b191_1_9_1.img",
    "image0001.pgm",
    "mgzn-20hpt.img",
    "oPPA_5grains_0001.tif",
    "XSDataImage.xml",
]

setup = """
import fabio
"""
stmt = "data = fabio.open(r'%s').data"


def run_benchmark(number=10, repeat=3):
    """
    :param number: Measure timimg over number of executions
    :param repeat: number of measurement, takes the best of them

    """
    print(f"Benchmarking during {number} seconds (best of {repeat} iterations).")
    print(f"Python {sys.version}")
    print(f"FabIO {version} ({date})")
    print("#" * 80)
    print(
        "     Module           filename        \t file size \t image size \t read time (ms) \t ms/Mpix"
    )
    for img in datasets:
        fn = utilstest.UtilsTest.getimage(img)
        fimg = fabio_open(fn)
        file_size = os.stat(fn).st_size / 1.0e6  # MB
        img_size = fimg.data.size / 1.0e6  # Mpix
        timer = timeit.Timer(stmt % fn, setup + (stmt % fn))
        tmin = min(
            [i / (0.001 * number) for i in timer.repeat(repeat=repeat, number=number)]
        )
        print(
            "%13s %25s %.3f Mb \t %.3f Mpix \t  %.3f ms \t %.3f ms/Mpix"
            % (fimg.__class__.__name__, img, file_size, img_size, tmin, tmin / img_size)
        )


run = run_benchmark



#!/usr/bin/env python
# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation files
#  (the "Software"), to deal in the Software without restriction,
#  including without limitation the rights to use, copy, modify, merge,
#  publish, distribute, sublicense, and/or sell copies of the Software,
#  and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

"""Portable image converter based on FabIO library
to export Eiger frames (including the one from LImA)
to a set of esperanto frames which can be imported
into CrysalisPro.
"""

__author__ = "Jérôme Kieffer"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__licence__ = "MIT"
__date__ = "27/10/2025"
__status__ = "production"


import logging
import sys
import os
import shutil
from .. import esperantoimage, eigerimage, limaimage, sparseimage, xcaliburimage
from ..openimage import openimage as fabio_open
from .. import version as fabio_version
from ..nexus import get_isotime
from ..utils.cli import ProgressBar, expand_args
import numpy
import argparse

try:
    import hdf5plugin  # noqa
except ImportError:
    pass

logging.basicConfig()
logger = logging.getLogger("eiger2crysalis")

try:
    import numexpr
except ImportError:
    logger.error("Numexpr is needed to interpret formula ...")

FOOTER = """To import your files as a project:
* Start CrysalisPro and open any project
* press "F5" to open the console
* Type `esperanto createrunlist` and select your first and last frame
"""

EXIT_SUCCESS = 0
EXIT_FAILURE = 1
EXIT_ARGUMENT_FAILURE = 2

try:
    from scipy import constants
except ImportError:
    # Updated with scipy 1.4
    CONST_hc = 12.398419843320026
else:
    CONST_hc = constants.c * constants.h / constants.e * 1e7


def as_str(smth):
    "Transform to string"
    if isinstance(any, bytes):
        return smth.decode()
    else:
        return str(smth)


class Converter:
    def __init__(self, options):
        self.options = options
        self.mask = None
        if not self.options.verbose:
            self.progress = ProgressBar("HDF5 --> Esperanto", len(options.images), 30)
        self.succeeded = True

        prefix = os.path.commonprefix([os.path.abspath(i) for i in self.options.images])
        if "{dirname}" in self.options.output:
            self.dirname = os.path.dirname(prefix)
        else:
            self.dirname = os.path.dirname(os.path.abspath(self.options.output))
        if "{prefix}" in self.options.output:
            self.prefix = os.path.basename(prefix) + "_1"
        else:
            self.prefix = os.path.basename(os.path.abspath(self.options.output)).split(
                "{"
            )[0]
        self.headers = None
        self.processed_frames = None
        self.scan_type = None
        self.angle_ranges = {}

    def geometry_transform(self, image):
        "Transforms an image according to the requested command line options"
        if self.options.rotation:
            image = numpy.rot90(image, k=self.options.rotation // 90)
        if self.options.transpose:
            image = image.T
        if self.options.flip_ud:
            image = numpy.flipud(image)
        if self.options.flip_lr:
            image = numpy.fliplr(image)
        return image

    def new_beam_center(self, x, y, shape):
        """Calculate the position of the beam after all transformations:

        :param x, y: position in the initial image
        :shape: shape of the input image
        :return: x, y, coordinates of the new beam center within the esperanto frame.
        """
        dummy = 123
        m = numpy.zeros(shape, dtype=numpy.int32)
        m[int(y + 0.5), int(x + 0.5)] = dummy

        f = esperantoimage.EsperantoImage(data=m)
        n = self.geometry_transform(f.data)
        w = numpy.argmin(abs(n.ravel() - dummy))
        return w % n.shape[-1], w // n.shape[-1]

    def convert_all(self):
        self.succeeded = True
        self.processed_frames = 0
        self.headers = self.common_headers()
        for filename in self.options.images:
            finish_at = self.convert_one(filename, self.processed_frames)
            self.succeeded = self.succeeded and (finish_at > 0)
            self.processed_frames += finish_at

    def finish(self):
        if not self.succeeded:
            if not self.options.verbose:
                self.progress.clear()
            print(
                "Conversion or part of it failed. You can try with --debug to have more output information."
            )
            return EXIT_FAILURE
        else:
            if not self.options.verbose:
                self.progress.clear()
            print(FOOTER)
            return EXIT_SUCCESS

    def common_headers(self):
        headers = {
            # SPECIAL_CCD_1
            "delectronsperadu": 1,
            "ldarkcorrectionswitch": 0,
            "lfloodfieldcorrectionswitch/mode": 0,
            "dsystemdcdb2gain": 1.0,
            "ddarksignal": 0,
            "dreadnoiserms": 0,
            # SPECIAL_CCD_2
            "ioverflowflag": 0,
            "ioverflowafterremeasureflag": 0,
            "inumofdarkcurrentimages": 0,
            "inumofmultipleimages": 0,
            "loverflowthreshold": 1000000,
            # SPECIAL_CCD_3
            # SPECIAL_CCD_4
            # SPECIAL_CCD_5
            # TIME
            # "dexposuretimeinsec": 0.2,
            "doverflowtimeinsec": 0,
            "doverflowfilter": 0,
            # MONITOR
            # PIXELSIZE
            # "drealpixelsizex": 0.075,
            # "drealpixelsizey": 0.075,
            "dsithicknessmmforpixeldetector": 1,
            # TIMESTAMP
            "timestampstring": get_isotime(),
            # GRIDPATTERN
            # STARTANGLESINDEG
            #             "dom_s":-180 + i,
            #             "dth_s":0,
            #             "dka_s":0,
            #             "dph_s":0,
            # ENDANGLESINDEG
            #             "dom_e":-179 + i,
            #             "dth_e": 0,
            #             "dka_e": 0,
            #             "dph_e": 0,
            # GONIOMODEL_1
            "dbeam2indeg": 0,
            "dbeam3indeg": 0,
            "detectorrotindeg_x": 0,
            "detectorrotindeg_y": 0,
            "detectorrotindeg_z": 0,
            #             "dxorigininpix":  img.data.shape[1] - (img.data.shape[1] - data.shape[1]) / 2 - center_x,
            #             "dyorigininpix": img.data.shape[0] - center_y,
            "dalphaindeg": 50,
            "dbetaindeg": 0,
            #                 "ddistanceinmm": 117,
            # GONIOMODEL_2
            # WAVELENGTH
            # "dalpha1": wl,
            # "dalpha2": wl,
            # "dalpha12": wl,
            # "dbeta1": wl,
            # MONOCHROMATOR
            "ddvalue-prepolfac": 0.98,
            "orientation-type": "SYNCHROTRON",
            # ABSTORUN
        }

        with fabio_open(self.options.images[0]) as source:
            shape = source.data.shape
            dtype = source.data.dtype
            if self.progress is not None:
                self.progress.max_value = source.nframes * len(self.options.images)
            if isinstance(source, limaimage.LimaImage):
                # Populate the Pilatus header from the Lima
                entry_name = source.h5.attrs.get("default")
                if entry_name:
                    entry = source.h5.get(entry_name)
                    if entry:
                        data_name = entry.attrs["default"]
                        if data_name:
                            data_grp = entry.get(data_name)
                            if data_grp:
                                nxdetector = data_grp.parent
                                try:
                                    headers["drealpixelsizex"] = (
                                        nxdetector[
                                            "detector_information/pixel_size/xsize"
                                        ][()]
                                        * 1e3
                                    )
                                    headers["drealpixelsizey"] = (
                                        nxdetector[
                                            "detector_information/pixel_size/ysize"
                                        ][()]
                                        * 1e3
                                    )
                                except Exception as e:
                                    logger.warning(
                                        "Error in searching for pixel size (%s): %s",
                                        type(e),
                                        e,
                                    )
                                try:
                                    t1 = nxdetector["acquisition/exposure_time"][()]
                                    headers["dexposuretimeinsec"] = t1
                                except Exception as e:
                                    logger.warning(
                                        "Error in searching for exposure time (%s): %s",
                                        type(e),
                                        e,
                                    )
            elif isinstance(source, sparseimage.SparseImage):
                entry_name = source.h5.attrs.get("default")
                if entry_name:
                    entry = source.h5.get(entry_name)
                    if entry:
                        instruments = [
                            i
                            for i in entry.values()
                            if as_str(i.attrs.get("NX_class", "")) == "NXinstrument"
                        ]
                        if instruments:
                            instrument = instruments[0]
                            detectors = [
                                i
                                for i in instrument.values()
                                if as_str(i.attrs.get("NX_class", "")) == "NXdetector"
                            ]
                            if detectors:
                                detector = detectors[0]
                                headers["drealpixelsizex"] = (
                                    detector["x_pixel_size"][()] * 1e3
                                )
                                headers["drealpixelsizey"] = (
                                    detector["y_pixel_size"][()] * 1e3
                                )
                                headers["dxorigininpix"] = detector["beam_center_x"][()]
                                headers["dyorigininpix"] = detector["beam_center_y"][()]
                                headers["ddistanceinmm"] = (
                                    detector["distance"][()] * 1e3
                                )
                            monchromators = [
                                i
                                for i in instrument.values()
                                if as_str(i.attrs.get("NX_class", ""))
                                == "NXmonochromator"
                            ]
                            if monchromators:
                                wavelength = monchromators[0]["wavelength"][()]
                self.mask = numpy.logical_not(numpy.isfinite(source.mask))
                headers["dexposuretimeinsec"] = 1  # meaningfull value.

            elif isinstance(source, eigerimage.EigerImage):
                entry_name = source.h5.attrs.get("default")
                if entry_name is None:
                    entry_name = "entry"
                entry = source.h5.get(entry_name)
                try:
                    nxdetector = entry["instrument/detector"]
                except Exception:
                    logger.error(
                        "No detector definition in Eiger file, is this a master file ?"
                    )
                else:
                    headers["drealpixelsizex"] = nxdetector["x_pixel_size"][()] * 1e3
                    headers["drealpixelsizey"] = nxdetector["y_pixel_size"][()] * 1e3
                    headers["dxorigininpix"] = nxdetector["beam_center_x"][()]
                    headers["dyorigininpix"] = nxdetector["beam_center_y"][()]
                    headers["ddistanceinmm"] = nxdetector["detector_distance"][()] * 1e3
                    headers["dexposuretimeinsec"] = nxdetector["count_time"][()]
                    wavelength = entry["instrument/beam/incident_wavelength"][()]
            else:
                raise NotImplementedError(
                    "Unsupported format: %s" % source.__class__.__name__
                )
        if self.mask is None:
            self.mask = numpy.zeros(shape, dtype=dtype)
        # Parse option for headers
        if self.options.energy:
            wavelength = CONST_hc / self.options.energy
        elif self.options.wavelength:
            wavelength = self.options.wavelength
        headers["dalpha1"] = headers["dalpha2"] = headers["dalpha12"] = headers[
            "dbeta1"
        ] = wavelength
        if self.options.distance:
            headers["ddistanceinmm"] = self.options.distance
        if self.options.beam:
            x, y = self.options.beam
            x, y = self.new_beam_center(x, y, shape)
            headers["dxorigininpix"] = x
            headers["dyorigininpix"] = y
        if self.options.alpha:
            headers["dalphaindeg"] = self.options.alpha
        if self.options.kappa is not None:
            try:
                value = float(self.options.kappa)
            except ValueError:  # Handle the string
                value = numexpr.NumExpr(self.options.kappa)
                self.scan_type = "kappa"
            headers["dka_s"] = headers["dka_e"] = value
        if self.options.theta is not None:
            try:
                value = float(self.options.theta)
            except ValueError:  # Handle the string
                value = numexpr.NumExpr(self.options.theta)
                self.scan_type = "theta"
            headers["dth_s"] = headers["dth_e"] = value
        if self.options.phi is not None:
            try:
                value = float(self.options.phi)
            except ValueError:  # Handle the string
                value = numexpr.NumExpr(self.options.phi)
                self.scan_type = "phi"
            headers["dph_s"] = headers["dph_e"] = value
        if self.options.omega is not None:
            try:
                value = float(self.options.omega)
            except ValueError:
                # Handle the string
                value = numexpr.NumExpr(self.options.omega)
                self.scan_type = "omega"
            headers["dom_s"] = headers["dom_e"] = value

        return headers

    def convert_one(self, input_filename, start_at=0):
        """
        Convert a single file using options

        :param str input_filename: The input filename
        :param object options: List of options provided from the command line
        :param start_at: index to start at for given file
        :rtype: int
        :returns: the number of frames processed
        """
        self.progress.update(start_at + 0.5, input_filename)
        input_filename = os.path.abspath(input_filename)
        input_exists = os.path.exists(input_filename)

        if self.options.verbose:
            print("Converting file '%s'" % (input_filename))

        if not input_exists:
            logger.error(
                "Input file '%s' do not exists. Conversion skipped.", input_filename
            )
            return -1

        try:
            logger.debug("Load '%s'", input_filename)
            source = fabio_open(input_filename)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            logger.error(
                "Loading input file '%s' failed cause: \"%s\". Conversion skipped.",
                input_filename,
                e.message,
            )
            logger.debug("Backtrace", exc_info=True)
            return -1

        for i, frame in enumerate(source):
            idx = i + start_at
            self.progress.update(idx + 0.5, input_filename + " - " + str(idx))
            input_data = frame.data
            numpy.maximum(self.mask, input_data, out=self.mask)
            input_data = input_data.astype(numpy.int32)
            input_data[input_data == numpy.iinfo(frame.data.dtype).max] = (
                self.options.dummy
            )
            converted = esperantoimage.EsperantoImage(
                data=input_data
            )  # This changes the shape
            converted.data = self.geometry_transform(converted.data)
            for k, v in self.headers.items():
                if callable(v):
                    if k.endswith("s"):
                        v0 = converted.header[k] = v(idx)
                    else:  # k.endswith("e"):
                        v1 = converted.header[k] = v(idx + 1)
                else:
                    v0 = v1 = converted.header[k] = v
                if k in self.angle_ranges:
                    v = self.angle_ranges[k]
                    self.angle_ranges[k] = (min(v[0], v0, v1), max(v[1], v0, v1))
                else:
                    self.angle_ranges[k] = (min(v0, v1), max(v0, v1))

            output_filename = self.options.output.format(
                index=(idx + self.options.offset),
                prefix=self.prefix,
                dirname=self.dirname,
            )
            os.makedirs(os.path.dirname(output_filename), exist_ok=True)
            try:
                logger.debug("Write '%s'", output_filename)
                if not self.options.dry_run:
                    converted.write(output_filename)
            except KeyboardInterrupt:
                raise
            except Exception as e:
                logger.error(
                    "Saving output file '%s' failed cause: \"%s: %s\". Conversion skipped.",
                    output_filename,
                    type(e),
                    e,
                )
                logger.debug("Backtrace", exc_info=True)
                return -1
        return source.nframes

    def treat_mask(self, full=False):
        ":param full: complete/slow mask analysis"
        if self.progress:
            self.progress.update(self.progress.max_value - 1, "Generate mask")
        dummy_value = numpy.asarray(-numpy.ones(1), dtype=self.mask.dtype)[
            -1
        ]  # -1 in the given datatype
        mask = (self.mask == dummy_value).astype(numpy.int8)
        esperantoimage.EsperantoImage.DUMMY = 1
        new_mask = self.geometry_transform(
            esperantoimage.EsperantoImage(data=mask).data
        )
        esperantoimage.EsperantoImage.DUMMY = -1  # restore the class !
        if self.progress:
            self.progress.update(
                self.progress.max_value - 1, f"Decompose mask full={full}"
            )
        xci = xcaliburimage.XcaliburImage(data=new_mask)
        ccd = xci.decompose(full)
        if self.progress:
            self.progress.update(
                self.progress.max_value - 0.5, "Exporting mask as CCD/SET file"
            )
        dummy_filename = self.options.output.format(
            index=self.options.offset, prefix=self.prefix, dirname=self.dirname
        )
        dirname = os.path.dirname(dummy_filename)
        prefix = self.prefix.split("_")[0]
        numpy.save(os.path.join(dirname, prefix + "_mask.npy"), new_mask)
        ccd.save(os.path.join(dirname, prefix + ".ccd"))
        with open(os.path.join(dirname, prefix + ".set"), mode="wb") as maskfile:
            maskfile.write(b"#XCALIBUR SYSTEM\r\n")
            maskfile.write(b"#XCALIBUR SETUP FILE\r\n")
            maskfile.write(
                b"#*******************************************************************************************************\r\n"
            )
            maskfile.write(
                b"# CHIP CHARACTERISTICS e_19_020609.ccd         D A T E Wed-Sep-16-10-00-59-2009\r\n"
            )
            maskfile.write(b"# This program produces version 1.9\r\n")
            maskfile.write(
                b"#******************************************************************************************************\r\n"
            )
            maskfile.write(
                b"#THIS FILE IS USER READABLE - BUT SHOULD NOT BE TOUCHED BY THE USER\r\n"
            )
            maskfile.write(
                b"#ANY CHANGES TO THIS FILE WILL RESULT IN LOSS OF WARRANTY!\r\n"
            )
            maskfile.write(b"#CHIP IDCODE producer type serial\r\n")
            maskfile.write(b'CHIP IDCODE "n/a" "n/a" "n/a\r\n')
            maskfile.write(b"#CHIP TAPER producer type serial\r\n")
            maskfile.write(b'CHIP TAPER "" "" ""\r\n')
            maskfile.write(b"#ALL COORDINATES GO FROM 0 TO N-1\r\n")
            maskfile.write(
                b"#CHIP BADPOINT treatment options: IGNORE,REPLACE,AVERAGE\r\n"
            )
            maskfile.write(
                b"#CHIP BADPOINT x1x1 y1x1 treatment r1x1x1 r1y1x1 r2x1x1 r2y1x1\r\n"
            )
            maskfile.write(b"#CHIP BADPOINT 630 422 REPLACE 632 422 0 0\r\n")
            maskfile.write(b"#CHIP BADRECTANGLE xl xr yb yt\r\n")
            for r in ccd.pschipbadpolygon:
                maskfile.write(
                    f"CHIP BADRECTANGLE {r.iax[0]} {r.iax[1]} {r.iay[0]} {r.iay[1]}\r\n".encode()
                )
            for r in ccd.pschipbadpoint:
                maskfile.write(
                    f"CHIP BADPOINT {r.spt.ix} {r.spt.iy} IGNORE {r.spt.ix} {r.spt.iy} {r.spt.ix} {r.spt.iy}\r\n".encode()
                )
            maskfile.write(b"#END OF XCALIBUR CHIP CHARACTERISTICS FILE\r\n")
        # Make a backup as the original could be overwritten by Crysalis at import
        shutil.copyfile(
            os.path.join(dirname, prefix + ".set"),
            os.path.join(dirname, prefix + ".set.orig"),
        )

        # save the ".run" file
        rundescription = xcaliburimage.RunDescription(prefix, dirname, pssweep=[])
        if self.scan_type == "phi":
            iscantype = xcaliburimage.SCAN_TYPE.Phi.value
            dstart = self.headers["dph_s"](0)
            dend = self.headers["dph_s"](self.processed_frames)
            dphi = 0.0
            domega = self.headers["dom_s"]
        elif self.scan_type == "omega":
            iscantype = xcaliburimage.SCAN_TYPE.Omega.value
            dstart = self.headers["dom_s"](0)
            dend = self.headers["dom_s"](self.processed_frames)
            dphi = self.headers["dph_s"]
            domega = 0.0
        else:
            raise RuntimeError(
                "This program only supports Omega and Phi scans, please provide the scan type in the command-line. See the output of the `--help` options."
            )
        oscil = (dend - dstart) / self.processed_frames
        sweep = xcaliburimage.Sweep(
            0,
            iscantype,
            domega=domega,
            dtheta=self.headers["dth_s"],
            dkappa=self.headers["dka_s"],
            dphi=dphi,
            dstart=dstart,
            dend=dend,
            dwidth=oscil,
            dunknown2=0.0,
            iunknown3=self.processed_frames,
            iunknown4=0,
            iunknown5=self.processed_frames,
            iunknown6=0,
            dexposure=self.headers["dexposuretimeinsec"],
        )
        rundescription.pssweep.append(sweep)
        rundescription.save(os.path.join(dirname, prefix + ".run"))

        # Finally save the ".par" file
        xci.save_par(
            dirname,
            prefix,
            wavelength=self.options.wavelength,
            alpha=self.options.alpha,
            polarization=self.options.polarization,
            distance=self.options.distance,
            oscil=oscil,
            center_x=self.headers["dxorigininpix"],
            center_y=self.headers["dyorigininpix"],
        )


def main():
    epilog = """return codes: 0 means a success. 1 means the conversion
                contains a failure, 2 means there was an error in the
                arguments"""

    parser = argparse.ArgumentParser(
        prog="eiger2crysalis", description=__doc__, epilog=epilog
    )
    parser.add_argument("IMAGE", nargs="*", help="File with input images")
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=fabio_version,
        help="output version and exit",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        default=False,
        help="show information for each conversions",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        dest="debug",
        default=False,
        help="show debug information",
    )
    group = parser.add_argument_group("main arguments")
    group.add_argument(
        "-l",
        "--list",
        action="store_true",
        dest="list",
        default=None,
        help="show the list of available formats and exit",
    )
    group.add_argument(
        "-o",
        "--output",
        default="{dirname}/{prefix}/{prefix}_1_{index}.esperanto",
        type=str,
        help="output directory and filename template",
    )
    group.add_argument(
        "-O",
        "--offset",
        type=int,
        default=1,
        help="index offset, CrysalisPro likes indexes to start at 1, Python starts at 0",
    )
    group.add_argument(
        "-D",
        "--dummy",
        type=int,
        default=-1,
        help="Set masked values to this dummy value",
    )

    group = parser.add_argument_group("optional behaviour arguments")
    #     group.add_argument("-f", "--force", dest="force", action="store_true", default=False,
    #                        help="if an existing destination file cannot be" +
    #                        " opened, remove it and try again (this option" +
    #                        " is ignored when the -n option is also used)")
    #     group.add_argument("-n", "--no-clobber", dest="no_clobber", action="store_true", default=False,
    #                        help="do not overwrite an existing file (this option" +
    #                        " is ignored when the -i option is also used)")
    #     group.add_argument("--remove-destination", dest="remove_destination", action="store_true", default=False,
    #                        help="remove each existing destination file before" +
    #                        " attempting to open it (contrast with --force)")
    #     group.add_argument("-u", "--update", dest="update", action="store_true", default=False,
    #                        help="copy only when the SOURCE file is newer" +
    #                        " than the destination file or when the" +
    #                        " destination file is missing")
    #     group.add_argument("-i", "--interactive", dest="interactive", action="store_true", default=False,
    #                        help="prompt before overwrite (overrides a previous -n" +
    #                        " option)")
    group.add_argument(
        "--dry-run",
        dest="dry_run",
        action="store_true",
        default=False,
        help="do everything except modifying the file system",
    )
    group.add_argument(
        "--calc-mask",
        dest="calc_mask",
        default=False,
        action="store_true",
        help="Generate a fine mask from pixels marked as invalid. By default, only treats gaps",
    )

    group = parser.add_argument_group("Experimental setup options")
    group.add_argument(
        "-e",
        "--energy",
        type=float,
        default=None,
        help="Energy of the incident beam in keV",
    )
    group.add_argument(
        "-w",
        "--wavelength",
        type=float,
        default=None,
        help="Wavelength of the incident beam in Å",
    )
    group.add_argument(
        "-d",
        "--distance",
        type=float,
        default=None,
        help="Detector distance in millimeters",
    )
    group.add_argument(
        "-b",
        "--beam",
        nargs=2,
        type=float,
        default=None,
        help="Direct beam in pixels x, y",
    )
    group.add_argument(
        "-p",
        "--polarization",
        type=float,
        default=0.99,
        help="Polarization factor (0.99 by default on synchrotron)",
    )

    group = parser.add_argument_group("Goniometer setup")
    #     group.add_argument("--axis", type=str, default=None,
    #                        help="Goniometer angle used for scanning: 'omega', 'phi' or 'kappa'")
    group.add_argument(
        "--alpha",
        type=float,
        default=50,
        help="Goniometer angle alpha value in deg. Constant, angle between kappa/omega.",
    )
    group.add_argument(
        "--kappa",
        type=str,
        default=0,
        help="Goniometer angle kappa value in deg or formula f(index).",
    )
    #     group.add_argument("--chi", type=str, default=0,
    #                        help="Goniometer angle chi value in deg. or formula f(index).")
    group.add_argument(
        "--phi",
        type=str,
        default=0,
        help="Goniometer angle phi value in deg. or formula f(index). Inner-most rotation.",
    )
    group.add_argument(
        "--omega",
        type=str,
        default=0,
        help="Goniometer angle omega value in deg. or formula f(index). Outer-most rotation.",
    )
    group.add_argument(
        "--theta",
        type=str,
        default=0,
        help="Goniometer angle theta value in deg. or formula f(index). Tilt angle of the detector.",
    )

    group = parser.add_argument_group(
        "Image preprocessing (Important: applied in this order!)"
    )
    group.add_argument(
        "--rotation",
        type=int,
        default=180,
        help="Rotate the initial image by this value in degrees. Must be a multiple of 90°. By default 180 deg (flip_up with origin=lower and flip_lr because the image is seen from the sample).",
    )
    group.add_argument(
        "--transpose", default=False, action="store_true", help="Flip the x/y axis"
    )
    group.add_argument(
        "--flip-ud",
        dest="flip_ud",
        default=False,
        action="store_true",
        help="Flip the image upside-down",
    )
    group.add_argument(
        "--flip-lr",
        dest="flip_lr",
        default=False,
        action="store_true",
        help="Flip the image left-right",
    )

    try:
        args = parser.parse_args()

        if args.debug:
            logger.setLevel(logging.DEBUG)

        if args.list:
            print("Supported formats: LimaImage, EigerImage, SparseImage")
            return

        if len(args.IMAGE) == 0:
            raise argparse.ArgumentError(None, "No input file specified.")

        # the upper case IMAGE is used for the --help auto-documentation
        args.images = expand_args(args.IMAGE)
        args.images.sort()
    except argparse.ArgumentError as e:
        logger.error(e.message)
        logger.debug("Backtrace", exc_info=True)
        return EXIT_ARGUMENT_FAILURE
    esperantoimage.EsperantoImage.DUMMY = args.dummy
    converter = Converter(args)
    converter.convert_all()
    converter.treat_mask(full=args.calc_mask)
    return converter.finish()


if __name__ == "__main__":
    result = main()
    sys.exit(result)



# coding: utf-8
#
#    Project: Azimuthal integration
#             https://github.com/silx-kit/pyFAI
#
#    Copyright (C) 2015-2025 European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Test module for utils.shell module"""

__author__ = "valentin.valls@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "19/06/2026"
__status__ = "development"
__docformat__ = "restructuredtext"

import unittest
import logging
from pathlib import Path
from .utilstest import UtilsTest
from ..utils.testutils import  LoggingCounter
from ..utils.cli import ProgressBar, expand_args
from .. import eigerimage
logger = logging.getLogger(__name__)


class TestUtilShell(unittest.TestCase):
    def test_coverage(self):
        """
        test function coverage
        """
        progressbar = ProgressBar("aaa", 15.1, 10)
        progressbar.update(1, "toto")
        progressbar.update(10.5, "toto")
        progressbar.update(20, "toto")
        progressbar.clear()

    def test_expand_args(self):
        _ = expand_args(["*.tif", "*.edf", "*.cbf", "*.img"])


class TestMisc(unittest.TestCase):

    def test_no_atleat_typo_in_source(self):
        # Locate the source file directly
        src_file = Path(eigerimage.__file__)
        self.assertTrue(src_file.exists(), f"Source file not found: {src_file}")
        content = src_file.read_text(encoding="utf-8")
        self.assertTrue("atleat_2d" not in content, "Typo 'atleat_2d' still present in eigerimage.py")


class TestCli(unittest.TestCase):
    def test_ulimit(self):
        import fabio.utils.cli
        with LoggingCounter(fabio.utils.cli._logger) as lc:
            fabio.utils.cli.relax_ulimit()
            counters = lc.counters

        try:
            import resource
        except Exception:
            return
        try:
            soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        except Exception as err:
            logger.error(f"{type(err)}: {err}")
        if counters.get(logging.warning) == 0:
            self.assertEqual(soft, hard, "soft limits have been increased to hard ones")
        else:
            logger.warning("Test skipped as `relax_ulimit` emitted warnings")


def suite():
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loader(TestUtilShell))
    testsuite.addTest(loader(TestMisc))
    testsuite.addTest(loader(TestCli))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())
    UtilsTest.clean_up()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

"""Test binary images"""

import unittest
import os
import logging
import numpy
from fabio.openimage import openimage
from fabio.binaryimage import BinaryImage
from ..utilstest import UtilsTest

logger = logging.getLogger(__name__)


def make_file(name, shape, dtype):
    data = numpy.random.random(shape).astype(dtype)
    numpy.save(name, data)
    return data


class TestBinaryImage(unittest.TestCase):
    """basic test"""

    @classmethod
    def setUpClass(cls):
        cls.fn3 = os.path.join(UtilsTest.tempdir, "binary.npy")
        cls.shape = (99, 101)
        cls.dtype = "<f"
        cls.data = make_file(cls.fn3, cls.shape, cls.dtype)

    def test_read(self):
        """check we can read images from Eiger"""
        e = BinaryImage()
        e.read(
            self.fn3,
            self.shape[1],
            self.shape[0],
            offset=-1,
            bytecode=self.dtype[1],
            endian=self.dtype[0],
        )
        f = openimage(self.fn3)
        self.assertEqual(e.shape, f.shape)
        self.assertEqual(e.bpp, f.bpp, "bpp OK")
        print(self.fn3)
        self.assertEqual(abs(e.data - f.data).max(), 0, "data OK")

    def test_write(self):
        fn = os.path.join(UtilsTest.tempdir, "binary_write.h5")
        ary = numpy.random.randint(0, 100, size=self.shape)
        e = BinaryImage(data=ary)
        e.save(fn)
        self.assertTrue(os.path.exists(fn), "file exists")
        f = BinaryImage()
        f.read(
            fn,
            self.shape[1],
            self.shape[0],
            offset=0,
            bytecode=ary.dtype,
            endian=ary.dtype.str[0],
        )
        self.assertEqual(
            str(f.__class__.__name__), "BinaryImage", "Used the write reader"
        )
        self.assertEqual(self.shape, f.shape, "shape matches")
        self.assertEqual(abs(f.data - ary).max(), 0, "first frame matches")


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestBinaryImage))
    return testsuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) 2015 European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""New Cython version of cf_iomodule.c for preparing the migration to Python3"""

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__date__ = "27/10/2025"
__copyright__ = "2013, European Synchrotron Radiation Facility, Grenoble, France"

from libc.string cimport memcpy
from libc.stdio cimport fopen, FILE

import cython
import numpy
import os
import tempfile
import logging
logger = logging.getLogger(__name__)


CF_H = 1

CF_INIT_ROWS = 8192
CF_INIT_COLS = 32
CF_HEADER_ITEM = 128


CF_GZ_COMP = 1
CF_BIN = 2


cdef extern from "columnfile.h":
    struct cf_data:
        int ncols, nrows
        unsigned int nralloc
        double **data
        char **clabels
    void * cf_read_ascii(void *fp, void *dest, unsigned int FLAGS)nogil
    void * cf_read_bin(void *fp, void *dest, unsigned int FLAGS)nogil
    int cf_write(char *fname, void *cf_handle, unsigned int FLAGS)nogil
    int cf_write_bin(void *fp, void *cf_handle)nogil
    int cf_write_ascii(void *fp, void *cf_handle,unsigned int FLAGS)nogil
    void cf_free( cf_data *cf_handle)nogil


def read(py_file, mode="a"):
    """
    Call the c-columnfile reading interface.
    The mode keyword argument is either:
    "a" for ascii (the default)
    "b" for binary
    """
    cdef cf_data *cf__
    cdef unsigned int flags = 0, fd

    # perhaps not const
    cdef int i
    cdef FILE *file

    # Here is a big issue !!! and I got an even worse solution !
    # file = PyFile_AsFile(py_file)
    (fd, fname) = tempfile.mkstemp()
    os.fdopen(fd, mode="wb").write(py_file.read())
    os.close(fd)
    file = fopen(fname, "r")

    if "z" in mode:
        flags |= CF_GZ_COMP
    if "b" in mode:
        cf__ = <cf_data *> cf_read_bin(file, NULL, flags)
    elif "a" in mode:
        cf__ =  <cf_data *> cf_read_ascii(file, NULL, flags)
    else:
        logger.error("Unrecognized mode for columnfile %s (assuming ascii)", mode)
        cf__ = <cf_data *> cf_read_ascii(file, NULL, flags)

    # check for failure to read
    if (cf__ == NULL):
        return None, None
    dims = (cf__.nrows, cf__.ncols)

    # since data may be non-contigous we can't simply create a numpy-array from
    # cf__->data, as Numpy's memory model prohibits it
    # i.e. py_data=(PyArrayObject*)PyArray_SimpleNewFromData(2, dims, NPY_DOUBLE, (void*)(&(cf__->data[0][0])))
    # won't work
    cdef double[:, ::1] py_data = numpy.empty(dims, dtype=numpy.float64)
    for i in range(cf__.nrows):
        memcpy(&py_data[i, 0], cf__.data[i], cf__.ncols * sizeof(double))
    clabels = []
    for i in range(cf__.ncols):
        clabels.append(str(cf__.clabels[i]))
    cf_free(cf__)

    return numpy.asarray(py_data), clabels
